"""Class for filtering prediction times before they are used for feature
generation."""
import logging
from typing import Optional

import pandas as pd

log = logging.getLogger(__name__)


class PredictionTimeFilterer:
    """Class for filtering prediction times before they are used for
    feature."""

    def __init__(
        self,
        prediction_times_df: pd.DataFrame,
        entity_id_col_name: str,
        quarantine_timestamps_df: Optional[pd.DataFrame] = None,
        quarantine_interval_days: Optional[int] = None,
        timestamp_col_name: Optional[str] = "timestamp",
    ):
        """Initialize PredictionTimeFilterer.

        Args:
            prediction_times_df (pd.DataFrame): Prediction times dataframe.
                Should contain entity_id and timestamp columns with col_names matching those in project_info.col_names.
            quarantine_df (pd.DataFrame, optional): A dataframe with timestamp column from which to start the quarantine.
                Any prediction times within the quarantine_interval_days after this timestamp will be dropped.
            quarantine_days (int, optional): Number of days to quarantine.
            entity_id_col_name (str): Name of the entity_id_col_name column.
            timestamp_col_name (str, optional): Name of the timestamp column.
        """

        self.prediction_times_df = prediction_times_df
        self.quarantine_df = quarantine_timestamps_df
        self.quarantine_days = quarantine_interval_days
        self.quarantine_df = self.quarantine_df.rename(
            columns={"timestamp": "timestamp_quarantine"},
        )
        self.entity_id_col_name = entity_id_col_name
        self.timestamp_col_name = timestamp_col_name

        self.added_pred_time_uuid_col: bool = False
        self.pred_time_uuid_col_name = "pred_time_uuid"

        uuid_cols = [c for c in self.prediction_times_df.columns if "uuid" in c]

        if len(uuid_cols) == 0:
            self.added_pred_time_uuid_col = True

            self.prediction_times_df[
                self.pred_time_uuid_col_name
            ] = self.prediction_times_df[self.entity_id_col_name].astype(
                str,
            ) + self.prediction_times_df[
                timestamp_col_name
            ].dt.strftime(
                "-%Y-%m-%d-%H-%M-%S",
            )

    def _filter_prediction_times_by_quarantine_period(self):
        # We need to check if ANY quarantine date hits each prediction time.
        # Create combinations
        n_before = len(self.prediction_times_df)

        df = self.prediction_times_df.merge(
            self.quarantine_df,
            on=self.entity_id_col_name,
            how="left",
        )

        df["days_since_quarantine"] = (
            df[self.timestamp_col_name] - df["timestamp_quarantine"]
        ).dt.days

        # Check if the prediction time is hit by the quarantine date.
        df.loc[
            (df["days_since_quarantine"] < self.quarantine_days)
            & (df["days_since_quarantine"] > 0),
            "hit_by_quarantine",
        ] = True

        # Get only the rows that were hit by the quarantine date
        df_hit_by_quarantine = df.loc[
            df["hit_by_quarantine"] == True  # pylint: disable=singleton-comparison
        ].drop_duplicates(subset=[self.pred_time_uuid_col_name])[
            ["pred_time_uuid", "hit_by_quarantine"]
        ]

        # Use these rows to filter the prediction times
        df = self.prediction_times_df.merge(
            df_hit_by_quarantine,
            on=self.pred_time_uuid_col_name,
            how="left",
            suffixes=("", "_hit_by_quarantine"),
            validate="one_to_one",
        )

        df = df.loc[
            df["hit_by_quarantine"] != True  # pylint: disable=singleton-comparison
        ]

        # Drop the columns we added
        df = df.drop(
            columns=[
                "hit_by_quarantine",
            ],
        )

        n_after = len(df)
        log.info(
            f"Filtered {n_before - n_after} prediction times by quarantine period.",
        )

        return df

    def filter(self):
        """Run filters based on the provided parameters."""
        df = self.prediction_times_df

        if self.quarantine_df is not None or self.quarantine_days is not None:
            if all([v is None for v in (self.quarantine_days, self.quarantine_df)]):
                raise ValueError(
                    "If either of quarantine_df and quarantine_days are provided, both must be provided.",
                )

            df = self._filter_prediction_times_by_quarantine_period()

        if self.added_pred_time_uuid_col:
            df = df.drop(columns=[self.pred_time_uuid_col_name])

        return df



"""Example of how to load demographic features."""
import pandas as pd

from psycop_feature_generation.loaders.raw.load_structured_sfi import (
    bmi,
    height_in_cm,
    weight_in_kg,
)
from psycop_feature_generation.loaders.raw.sql_load import sql_load

if __name__ == "__main__":
    df = sql_load(
        query="SELECT * FROM [fct].[FOR_SFI_vaegt_hoejde_BMI_psyk_somatik_inkl_2021]",
        database="USR_PS_FORSK",
        chunksize=None,
        n_rows=1_000,
    )[["aktivitetstypenavn", "elementledetekst", "numelementvaerdi", "elementvaerdi"]]

    df_bmi = bmi(n_rows=1_000)
    df_height = height_in_cm(n_rows=100_000)
    df_weight = weight_in_kg(n_rows=100_000)


def unique_and_percentage(series: pd.Series) -> pd.Series:
    """Return unique values and their percentage of the total number of values
    in the series.

    Args:
        series (pd.Series): Series to get unique values and percentage of.

    Returns:
        pd.Series: Series with unique values as index and percentage as values.
    """
    unique_values = series.value_counts(normalize=True)
    return unique_values



"""Import all raw loaders."""

from ...data_checks.raw.check_predictor_lists import (  # noqa
    check_feature_combinations_return_correct_dfs,
)
from .load_coercion import *  # noqa
from .load_demographic import *  # noqa
from .load_diagnoses import *  # noqa
from .load_ids import *  # noqa
from .load_lab_results import *  # noqa
from .load_medications import *  # noqa
from .load_structured_sfi import *  # noqa
from .load_t2d_outcomes import *  # noqa
from .load_visits import *  # noqa
from .sql_load import *  # noqa



"""Describe flattened dataset.""" ""
import logging

from psycop_feature_generation.application_modules.project_setup import ProjectInfo
from psycop_feature_generation.application_modules.wandb_utils import (
    wandb_alert_on_exception,
)
from psycop_feature_generation.data_checks.flattened.data_integrity import (
    save_feature_set_integrity_from_dir,
)

log = logging.getLogger(__name__)


@wandb_alert_on_exception
def save_flattened_dataset_description_to_disk(
    project_info: ProjectInfo,
    describe_splits: bool = True,
    compare_splits: bool = True,
):
    """Describe output.

    Args:
        project_info (ProjectInfo): Project info
        describe_splits (bool, optional): Whether to describe each split. Defaults to True.
        compare_splits (bool, optional): Whether to compare splits, e.g. do all categories exist in both train and val. Defaults to True.
    """
    feature_set_description_path = (
        project_info.feature_set_path / "feature_set_description"
    )
    split_feature_distribution_comparison_path = (
        project_info.feature_set_path / "split_feature_distribution_comparison_path"
    )

    log.info(
        f"Saving flattened dataset description to disk. Check {feature_set_description_path} and {split_feature_distribution_comparison_path} to validate that your dataset is not broken in some way.",
    )

    for method in ("describe", "compare"):
        # Describe the feature set
        if method == "describe" and describe_splits:
            splits = ["train"]

            compare_splits_in_method = False
            out_dir = feature_set_description_path

        # Compare train, val and test, to make sure they have the same categories
        if method == "compare" and compare_splits:
            splits = ["train", "val", "test"]

            # Don't describe specs in comparison, to avoid leakage by describing features in val and test
            compare_splits_in_method = True
            out_dir = split_feature_distribution_comparison_path

        save_feature_set_integrity_from_dir(
            splits=splits,
            out_dir=out_dir,
            dataset_format=project_info.dataset_format,
            describe_splits=describe_splits,
            compare_splits=compare_splits_in_method,
            feature_set_dir=project_info.feature_set_path,
        )



"""Feature loaders for loading .csv from disk."""

from pathlib import Path
from typing import Optional

import pandas as pd

from psycop_feature_generation.utils import load_dataset_from_file


def get_predictors(df: pd.DataFrame, include_id: bool) -> pd.DataFrame:
    """Returns the predictors from a dataframe.

    Assumes predictors to be prefixed with 'pred'. Timestamp is also
    returned for predictors, and optionally dw_ek_borger.

    Args:
        df: The dataframe to get the predictors from
        include_id (bool): Whether to include 'dw_ek_borger' in the returned df

    Returns:
        pd.DataFrame: Dataframe with only predictor columns
    """
    pred_regex = (
        "^pred|^timestamp" if not include_id else "^pred|^timestamp|dw_ek_borger"
    )
    return df.filter(regex=pred_regex)


def load_split(
    feature_set_dir: Path,
    file_suffix: str,
    split: str,
    nrows: Optional[int] = None,
) -> pd.DataFrame:
    """Loads a given data split as a dataframe from a directory.

    Args:
        feature_set_dir (Path): Path to directory containing data files
        file_suffix (str): The suffix of the files to load.
        split (str): Which string to look for (e.g. 'train', 'val', 'test')
        nrows (Optional[int]): Whether to only load a subset of the dat

    Returns:
        pd.DataFrame: The loaded dataframe
    """
    file_path = list(feature_set_dir.glob(f"*{split}*{file_suffix}"))[0]

    return load_dataset_from_file(file_path=file_path, nrows=nrows)


def load_split_predictors(
    feature_set_dir: Path,
    file_suffix: str,
    split: str,
    include_id: bool,
    nrows: Optional[int] = None,
) -> pd.DataFrame:
    """Loads predictors from a given data split as a dataframe from a
    directory.

    Args:
        feature_set_dir (Path): Path to directory containing data files
        file_suffix (str): The suffix of the files to load.
        split (str): Which string to look for (e.g. 'train', 'val', 'test')
        include_id (bool): Whether to include 'dw_ek_borger' in the returned df
        nrows (Optional[int]): Whether to only load a subset of the data

    Returns:
        pd.DataFrame: The loaded dataframe
    """

    return get_predictors(
        load_split(
            feature_set_dir=feature_set_dir,
            file_suffix=file_suffix,
            split=split,
            nrows=nrows,
        ),
        include_id,
    )


def get_outcomes(df: pd.DataFrame) -> pd.DataFrame:
    """Returns the outcomes from a dataframe.

    Assumes outcomes to be prefixed with 'outc'.

    Args:
        df: The dataframe to get the outcomes from

    Returns:
        pd.DataFrame: Dataframe with only outcome columns
    """
    return df.filter(regex="^outc")


def load_split_outcomes(
    feature_set_dir: Path,
    file_suffix: str,
    split: str,
    nrows: Optional[int] = None,
) -> pd.DataFrame:
    """Loads outcomes from a given data split as a dataframe from a directory.

    Args:
        feature_set_dir (Path): Path to directory containing data files
        file_suffix (str): The suffix of the files to load.
        split (str): Which string to look for (e.g. 'train', 'val', 'test')
        nrows (Optional[int]): Whether to only load a subset of the data

    Returns:
        pd.DataFrame: The loaded dataframe
    """
    return get_outcomes(
        load_split(
            feature_set_dir=feature_set_dir,
            split=split,
            nrows=nrows,
            file_suffix=file_suffix,
        ),
    )



"""Code to generate data integrity and train/val/test drift reports."""
from collections.abc import Iterable
from pathlib import Path
from typing import Any, Optional

import pandas as pd
from deepchecks.core.suite import SuiteResult
from deepchecks.tabular import Dataset, Suite
from deepchecks.tabular.checks import (
    CategoryMismatchTrainTest,
    DataDuplicates,
    DatasetsSizeComparison,
    FeatureFeatureCorrelation,
    FeatureLabelCorrelation,
    FeatureLabelCorrelationChange,
    IdentifierLabelCorrelation,
    IndexTrainTestLeakage,
    IsSingleValue,
    MixedDataTypes,
    MixedNulls,
    NewLabelTrainTest,
    OutlierSampleDetection,
    StringLengthOutOfBounds,
    TrainTestLabelDrift,
)
from wasabi import Printer

from psycop_feature_generation.loaders.flattened import (
    load_split_outcomes,
    load_split_predictors,
)


def pruned_data_integrity_checks(**kwargs) -> Suite:
    """Deepchecks data integrity suite with only wanted checks.
    Disables: SpecialCharacters, StringMismatch, ConflictingLabels.

    Args:
        **kwargs: keyword arguments to be passed to all checks.

    Returns:
        Suite: a deepchecks Suite
    """
    return Suite(
        "Data Integrity Suite",
        IsSingleValue(**kwargs).add_condition_not_single_value(),
        MixedNulls(**kwargs).add_condition_different_nulls_less_equal_to(),
        MixedDataTypes(**kwargs).add_condition_rare_type_ratio_not_in_range(),
        DataDuplicates(**kwargs).add_condition_ratio_less_or_equal(),
        StringLengthOutOfBounds(
            **kwargs
        ).add_condition_ratio_of_outliers_less_or_equal(),
        OutlierSampleDetection(**kwargs),
        FeatureLabelCorrelation(**kwargs).add_condition_feature_pps_less_than(),
        FeatureFeatureCorrelation(
            **kwargs
        ).add_condition_max_number_of_pairs_above_threshold(),
        IdentifierLabelCorrelation(**kwargs).add_condition_pps_less_or_equal(),
    )


def label_integrity_checks() -> Suite:
    """Deepchecks data integrity suite for checks that require a label.

    Returns:
        Suite: A deepchecks Suite

    Example:
    >>> suite = label_integrity_checks()
    >>> result = suite.run(some_deepchecks_dataset)
    >>> result.show()
    """
    return Suite(
        "Data integrity checks requiring labels",
        IdentifierLabelCorrelation().add_condition_pps_less_or_equal(),
        FeatureLabelCorrelation().add_condition_feature_pps_less_than(),
    )


def custom_train_test_validation(**kwargs) -> Suite:
    """Deepchecks train/test validation suite for train/test checks which slow
    checks disabled.

    Args:
        **kwargs: Keyword arguments to pass to the Suite constructor.

    Returns:
        Suite: A deepchecks Suite
    """
    return Suite(
        "Train Test Validation Suite",
        DatasetsSizeComparison(
            **kwargs
        ).add_condition_test_train_size_ratio_greater_than(),
        NewLabelTrainTest(**kwargs).add_condition_new_labels_number_less_or_equal(),
        CategoryMismatchTrainTest(
            **kwargs
        ).add_condition_new_category_ratio_less_or_equal(),
        IndexTrainTestLeakage(**kwargs).add_condition_ratio_less_or_equal(),
    )


def label_split_checks() -> Suite:
    """Deepchecks train/test validation suite for checks that require a label.

    Returns:
        Suite: a deepchecks Suite
    """
    return Suite(
        "Split validation checks requiring labels",
        FeatureLabelCorrelationChange()
        .add_condition_feature_pps_difference_less_than()
        .add_condition_feature_pps_in_train_less_than(),
        TrainTestLabelDrift().add_condition_drift_score_less_than(),
    )


def get_failed_check_names(result: SuiteResult) -> list[str]:
    """Returns a list of names of failed checks.

    Args:
        result (SuiteResult): A deepchecks SuiteResult

    Returns:
        list[str]: list of names of failed checks
    """
    return [
        check_result.check.name() for check_result in result.get_not_passed_checks()
    ]


def check_train_data_integrity(
    feature_set_dir: Path,
    file_suffix: str,
    out_dir: Path,
    train_outcomes_df: pd.DataFrame,
    outcome_checks_dir: Path,
    n_rows: Optional[int] = None,
):
    """Runs Deepcheck data integrity checks for the train split.

    Args:
        feature_set_dir (Path): Path to a directory containing train/val/test files
        file_suffix (str): File suffix of the train/val/test files
        out_dir (Path): Path to the directory where the reports should be saved
        train_outcomes_df (pd.DataFrame): The train outcomes dataframe
        outcome_checks_dir (Path): Path to the directory where the outcome specific reports should be saved
        n_rows (Optional[int]): Whether to only load a subset of the data.
            Should only be used for debugging.

    Returns:
        failures (dict): A dictionary containing the failed checks
    """
    failures = {}

    msg = Printer(timestamp=True)

    msg.info("Running data integrity checks...")

    # Only running data integrity checks on the training set to reduce the
    # chance of any form of peaking at the test set
    train_predictors = load_split_predictors(
        feature_set_dir=feature_set_dir,
        split="train",
        include_id=True,
        nrows=n_rows,
        file_suffix=file_suffix,
    )

    data_s = Dataset(
        df=train_predictors,
        index_name="dw_ek_borger",
        datetime_name="timestamp",
        cat_features=[],
    )

    # Running checks that do not require a label
    integ_suite = pruned_data_integrity_checks(
        timeout=0,
    )  # timeout=0 removes timeout

    suite_results = integ_suite.run(data_s)
    suite_results.save_as_html(str(out_dir / "data_integrity.html"))

    failures["data_integrity"] = get_failed_check_names(suite_results)

    # Running checks that require a label for each outcome
    label_checks = label_integrity_checks()

    for outcome_column in train_outcomes_df.columns:
        msg.info(f"Running data integrity for {outcome_column}")

        data_s = Dataset(
            df=train_predictors,
            index_name="dw_ek_borger",
            datetime_name="timestamp",
            label=train_outcomes_df[outcome_column],
            cat_features=[],
        )

        suite_results = label_checks.run(data_s)

        suite_results.save_as_html(
            str(outcome_checks_dir / f"{outcome_column}_check.html"),
        )

        failures[f"{outcome_column}_check"] = get_failed_check_names(
            suite_results,
        )

    msg.good("Finshed data integrity checks!")

    return failures


def get_suite_results_for_split_pair_and_save_to_disk(
    out_dir: Path,
    deepchecks_suite: Any,
    split_dicts: dict[str, dict[str, Any]],
    split_pair: tuple[str, str],
    file_suffix: str = "",
) -> SuiteResult:
    """Runs a Deepchecks suite on a given split and saves the results to a
    file.

    Args:
        out_dir (Path): Path to the directory where the results should be saved
        deepchecks_suite (Any): Deepchecks suite to run
        split_dicts (dict[str, dict[str, Any]]): Dictionary containing the splits.
        split_pair (tuple[str, str]): Splits to run the suite on
        file_suffix (str, optional): Suffix to add to the file name. Defaults to "".

    Returns:
        SuiteResults: Results of the suite
    """

    suite_results = deepchecks_suite.run(
        split_dicts[split_pair[0]]["ds"],
        split_dicts[split_pair[1]]["ds"],
    )

    suite_results.save_as_html(
        str(out_dir / f"{split_pair[0]}_{split_pair[1]}_{file_suffix}.html"),
    )

    return suite_results


def get_split_as_ds_dict(
    feature_set_dir: Path,
    n_rows: Optional[int],
    split: str,
    file_suffix: str,
) -> dict[str, Any]:
    """Loads a split as a Deepchecks Dataset dict.

    Args:
        feature_set_dir (Path): Path to a directory containing train/val/test files
        n_rows (Optional[int]): Whether to only load a subset of the data.
            Should only be used for debugging.
        split (str): Which split to load
        file_suffix (str, optional): Suffix of the files.

    Returns:
        dict: Deepchecks Dataset dict
    """

    predictors = load_split_predictors(
        feature_set_dir=feature_set_dir,
        split=split,
        include_id=True,
        nrows=n_rows,
        file_suffix=file_suffix,
    )
    outcomes = load_split_outcomes(
        feature_set_dir=feature_set_dir,
        split=split,
        nrows=n_rows,
        file_suffix=file_suffix,
    )

    data_s = Dataset(
        df=predictors,
        index_name="dw_ek_borger",
        datetime_name="timestamp",
        cat_features=[],
    )

    return {
        "predictors": predictors,
        "outcomes": outcomes,
        "ds": data_s,
    }


def run_validation_requiring_split_comparison(
    feature_set_dir: Path,
    splits: Iterable[str],
    file_suffix: str,
    out_dir: Path,
    train_outcome_df: pd.DataFrame,
    n_rows: Optional[int] = None,
):
    """Runs Deepcheck data validation checks for the train/val/test splits.

    Args:
        feature_set_dir (Path): Path to a directory containing train/val/test files
        splits (list[str]): list of splits to check (train, val, test)
        file_suffix (str): File suffix of the train/val/test files
        out_dir (Path): Path to the directory where the reports should be saved
        train_outcome_df (pd.DataFrame): The train outcomes.
        n_rows (int): Whether to only load a subset of the data.
            Should only be used for debugging.
    """
    msg = Printer(timestamp=True)

    failed_checks = {}

    validation_suite = custom_train_test_validation()

    split_dicts = {}

    for split_name in splits:
        split_dicts[split_name] = get_split_as_ds_dict(
            feature_set_dir=feature_set_dir,
            n_rows=n_rows,
            split=split_name,
            file_suffix=file_suffix,
        )

    for split_pair in (("train", "val"), ("train", "test")):
        suite_results = get_suite_results_for_split_pair_and_save_to_disk(
            out_dir=out_dir,
            deepchecks_suite=validation_suite,
            split_dicts=split_dicts,
            split_pair=split_pair,
            file_suffix="integrity",
        )

        failed_checks[
            f"{split_pair[0]}_{split_pair[1]}_integrity"
        ] = get_failed_check_names(suite_results)

    for split_name, split_contents in split_dicts.items():
        # don't check train/train
        if split_name == "train":
            continue

        for outcome_col in train_outcome_df:
            msg.info(
                f"Running split validation for train/{split_name} and {outcome_col}",
            )

            deepchecks_ds_dict = {
                "train": {
                    "ds": Dataset(
                        df=split_dicts["train"]["predictors"],
                        index_name="dw_ek_borger",
                        datetime_name="timestamp",
                        label=split_dicts["train"]["outcomes"][outcome_col],
                        cat_features=[],
                    ),
                },
                split_name: {
                    "ds": Dataset(
                        df=split_contents["predictors"],
                        index_name="dw_ek_borger",
                        datetime_name="timestamp",
                        label=split_contents["outcomes"][outcome_col],
                        cat_features=[],
                    ),
                },
            }

            suite_results = get_suite_results_for_split_pair_and_save_to_disk(
                out_dir=out_dir,
                deepchecks_suite=label_split_checks(),
                split_dicts=deepchecks_ds_dict,
                split_pair=("train", split_name),
                file_suffix=f"{outcome_col}_check",
            )

            failed_checks[
                f"train_{split_name}_{outcome_col}_check"
            ] = get_failed_check_names(suite_results)

        msg.good(f"All data checks done! Saved to {out_dir}")

        if len(failed_checks.keys()) > 0:
            msg.warn(f"Failed checks: {failed_checks}")


def save_feature_set_integrity_from_dir(  # noqa pylint: disable=too-many-statements
    feature_set_dir: Path,
    n_rows: Optional[int] = None,
    splits: Iterable[str] = ("train", "val", "test"),
    out_dir: Optional[Path] = None,
    dataset_format: str = "parquet",
    describe_splits: bool = True,
    compare_splits: bool = True,
) -> None:
    """Runs Deepcheck data integrity and train/val/test checks for a given
    directory containing train/val/test files. Splits indicates which data.
    splits to check.

    The resulting reports are saved to a sub directory as .html files.

    Args:
        feature_set_dir (Path): Path to a directory containing train/val/test files
        n_rows (Optional[int]): Whether to only load a subset of the data.
            Should only be used for debugging.
        splits (list[str]): list of splits to check (train, val, test)
        out_dir (Optional[Path]): Path to the directory where the reports should be saved
        dataset_format (str, optional): Format of the files to load. Must be either "csv" or "parquet". Defaults to "parquet".
        describe_splits (bool, optional): Whether to describe each split. Defaults to True.
        compare_splits (bool, optional): Whether to compare splits, e.g. do all categories exist in both train and val. Defaults to True.
    """
    if dataset_format not in ("parquet", "csv"):
        raise ValueError(
            f"file_suffix must be either 'parquet' or 'csv', got {dataset_format}",
        )

    if out_dir is None:
        out_dir = feature_set_dir / "deepchecks"
    else:
        out_dir = out_dir / "deepchecks"

    out_dir.mkdir(parents=True, exist_ok=True)

    train_outcomes_df = load_split_outcomes(
        feature_set_dir=feature_set_dir,
        split="train",
        nrows=n_rows,
        file_suffix=dataset_format,
    )

    failed_checks = (
        {}
    )  # Collect failed checks for error messages at the end of the function

    # Check if file splits exist before running checks
    for split_name in splits:
        file = list(feature_set_dir.glob(f"*{split_name}*{dataset_format}"))

        if not file:  # pylint: disable=consider-using-assignment-expr
            raise ValueError(f"{split_name} split not found in {feature_set_dir}")
        if len(file) > 1:
            raise ValueError(
                f"Multiple {split_name} files found in {feature_set_dir}",
            )

    # Create subfolder for outcome specific checks
    outcome_checks_dir = out_dir / "outcomes"
    if not outcome_checks_dir.exists():
        outcome_checks_dir.mkdir()

    if describe_splits:
        # Check train data integrity
        if "train" in splits:
            failures = check_train_data_integrity(
                feature_set_dir=feature_set_dir,
                n_rows=n_rows,
                out_dir=out_dir,
                outcome_checks_dir=outcome_checks_dir,
                train_outcomes_df=train_outcomes_df,
                file_suffix=dataset_format,
            )

            # Add all keys in failures to failed_checks
            for k, v in failures.items():
                failed_checks[k] = v

    if compare_splits:
        # Running data validation checks on train/val and train/test splits that do not
        # require a label
        run_validation_requiring_split_comparison(
            feature_set_dir=feature_set_dir,
            splits=splits,
            n_rows=n_rows,
            out_dir=out_dir,
            train_outcome_df=train_outcomes_df,
            file_suffix=dataset_format,
        )



"""Example for loading structured SFIs."""

import psycop_feature_generation.loaders.raw.load_structured_sfi as struct_sfi_loader
from psycop_feature_generation.data_checks.raw.check_predictor_lists import (
    check_feature_combinations_return_correct_dfs,
)

if __name__ == "__main__":
    df = struct_sfi_loader.selvmordsrisiko(n_rows=1000)

    input_dict = [
        {
            "predictor_df": "selvmordsrisiko",
            "allowed_nan_value_prop": 0.01,
        },
    ]

    check_feature_combinations_return_correct_dfs(
        predictor_dict_list=input_dict,
        n_rows=1000,
    )



"""Loaders for demographic information."""

from typing import Optional

import pandas as pd

from psycop_feature_generation.loaders.raw.sql_load import sql_load
from psycop_feature_generation.utils import data_loaders

# pylint: disable=missing-function-docstring


@data_loaders.register("birthdays")  # noqa
def birthdays(n_rows: Optional[int] = None) -> pd.DataFrame:
    view = "[FOR_kohorte_demografi_inkl_2021_feb2022]"

    sql = f"SELECT dw_ek_borger, foedselsdato FROM [fct].{view}"

    df = sql_load(sql, database="USR_PS_FORSK", chunksize=None, n_rows=n_rows)

    # Typically handled by sql_load, but because foedselsdato doesn't contain "datotid" in its name,
    # We must handle it manually here
    df["foedselsdato"] = pd.to_datetime(df["foedselsdato"], format="%Y-%m-%d")

    df.rename(columns={"foedselsdato": "date_of_birth"}, inplace=True)

    # msg.good("Loaded birthdays")
    return df.reset_index(drop=True)


@data_loaders.register("sex_female")
def sex_female(n_rows: Optional[int] = None) -> pd.DataFrame:
    view = "[FOR_kohorte_demografi_inkl_2021_feb2022]"

    sql = f"SELECT dw_ek_borger, koennavn FROM [fct].{view}"

    df = sql_load(sql, database="USR_PS_FORSK", chunksize=None, n_rows=n_rows)

    df.loc[df["koennavn"] == "Mand", "koennavn"] = False
    df.loc[df["koennavn"] == "Kvinde", "koennavn"] = True

    df.rename(
        columns={"koennavn": "sex_female"},
        inplace=True,
    )

    return df.reset_index(drop=True)



"""Check that all feature_dicts conform to correct formatting.

Also check that they return meaningful dictionaries.
"""
from typing import Optional, Union

import pandas as pd
from wasabi import Printer

from psycop_feature_generation.data_checks.raw.check_raw_df import check_raw_df
from psycop_feature_generation.utils import data_loaders


def check_df_conforms_to_feature_spec(
    df: pd.DataFrame,
    required_columns: list[str],
    subset_duplicates_columns: list[str],
    expected_val_dtypes: list[str],
    msg_prefix: str,
    arg_dict: dict,
    allowed_nan_value_prop: float = 0.01,
):
    """Check that a loaded df conforms to to a given feature specification.
    Useful when creating loaders.

    Args:
        df (pd.DataFrame): Dataframe to check.
        required_columns (list[str]): list of required columns.
        subset_duplicates_columns (list[str]): list of columns to subset on when
            checking for duplicates.
        expected_val_dtypes (list[str]): Expected value dtype.
        msg_prefix (str): Prefix to add to all messages.
        arg_dict (dict): Dictionary with specifications for what df should conform to.
        allowed_nan_value_prop (float): Allowed proportion of missing values. Defaults to 0.01.

    Raises:
        ValueError: If df does not conform to d.
    """

    if required_columns is None:
        required_columns = ["dw_ek_borger", "timestamp", "value"]

    if subset_duplicates_columns is None:
        subset_duplicates_columns = ["dw_ek_borger", "timestamp", "value"]

    if expected_val_dtypes is None:
        expected_val_dtypes = ["float64", "int64"]

    msg = Printer(timestamp=True)

    allowed_nan_value_prop = (
        arg_dict["allowed_nan_value_prop"]
        if arg_dict["allowed_nan_value_prop"]
        else allowed_nan_value_prop
    )

    source_failures, _ = check_raw_df(
        df=df,
        required_columns=required_columns,
        subset_duplicates_columns=subset_duplicates_columns,
        allowed_nan_value_prop=allowed_nan_value_prop,
        expected_val_dtypes=expected_val_dtypes,
    )

    # Return errors
    if len(source_failures) != 0:
        msg.fail(f"{msg_prefix} errors: {source_failures}")
        return {arg_dict["predictor_df"]: source_failures}
    else:
        msg.good(
            f"{msg_prefix} passed data validation criteria.",
        )
        return None


def get_predictor_df_with_loader_fn(d: dict, n_rows: int) -> pd.DataFrame:
    """Get predictor_df from d.

    Args:
        d (dict): Dictionary with key predictor_df
        n_rows (int): Number of rows to load

    Returns:
        pd.DataFrame: predictor_df
    """

    loader_fns_dict = data_loaders.get_all()

    if "loader_kwargs" in d:
        return loader_fns_dict[d["predictor_df"]](n_rows=n_rows, **d["loader_kwargs"])
    else:
        return loader_fns_dict[d["predictor_df"]](n_rows=n_rows)


def get_unique_predictor_dfs(predictor_dict_list: list[dict], required_keys: list[str]):
    """Get unique predictor_dfs from predictor_dict_list.

    Args:
        predictor_dict_list (list[dict]): list of dictionaries where the key predictor_df maps to a catalogue registered data loader
            or is a valid dataframe.
        required_keys (list[str]): list of required keys.

    Returns:
        list[dict]: list of unique predictor_dfs.
    """

    dicts_with_subset_keys = []

    for d in predictor_dict_list:
        new_d = {k: d[k] for k in required_keys}

        if "loader_kwargs" in d:
            new_d["loader_kwargs"] = d["loader_kwargs"]

        dicts_with_subset_keys.append(new_d)

    unique_subset_dicts = []

    for predictor_dict in dicts_with_subset_keys:
        if predictor_dict not in unique_subset_dicts:
            unique_subset_dicts.append(predictor_dict)

    return unique_subset_dicts


def check_feature_combinations_return_correct_dfs(  # noqa pylint: disable=too-many-branches
    predictor_dict_list: list[dict[str, Union[str, float, int]]],
    n_rows: int = 1_000,
    required_columns: Optional[list[str]] = None,
    subset_duplicates_columns: Optional[list[str]] = None,
    allowed_nan_value_prop: float = 0.01,
    expected_val_dtypes: Optional[list[str]] = None,
):
    """Test that all predictor_dfs in predictor_list return a valid df.

    Args:
        predictor_dict_list (list[dict[str, Union[str, float, int]]]): list of dictionaries
            where the key predictor_df maps to a catalogue registered data loader
            or is a valid dataframe.
        n_rows (int): Number of rows to test. Defaults to 1_000.
        required_columns (list[str]): list of required columns. Defaults to ["dw_ek_borger", "timestamp", "value"].
        subset_duplicates_columns (list[str]): list of columns to subset on when
            checking for duplicates. Defaults to ["dw_ek_borger", "timestamp"].
        allowed_nan_value_prop (float): Allowed proportion of missing values. Defaults to 0.0.
        expected_val_dtypes (list[str]): Expected value dtype. Defaults to ["float64", "int64"].
    """

    if required_columns is None:
        required_columns = ["dw_ek_borger", "timestamp", "value"]

    if subset_duplicates_columns is None:
        subset_duplicates_columns = ["dw_ek_borger", "timestamp", "value"]

    if expected_val_dtypes is None:
        expected_val_dtypes = ["float64", "int64"]

    msg = Printer(timestamp=True)

    msg.info("Checking that feature combinations conform to correct formatting")

    # Find all dicts that are unique on keys predictor_df and allowed_nan_value_prop
    required_keys = ["predictor_df", "allowed_nan_value_prop"]

    unique_subset_dicts = get_unique_predictor_dfs(
        predictor_dict_list=predictor_dict_list,
        required_keys=required_keys,
    )

    msg.info(f"Loading {n_rows} rows from each predictor_df")

    failure_dicts = []

    # Check each predictor df
    for i, d in enumerate(unique_subset_dicts):  # pylint: disable=invalid-name
        # Check that it returns a dataframe
        try:
            df = get_predictor_df_with_loader_fn(d=d, n_rows=n_rows)
        except KeyError:
            msg.warn(
                f"{d['predictor_df']} does not appear to be a loader function in catalogue, assuming a well-formatted dataframe. Continuing.",
            )
            continue

        msg_prefix = f"{i+1}/{len(unique_subset_dicts)} {d['predictor_df']}:"

        df_failures = check_df_conforms_to_feature_spec(
            df=df,
            required_columns=required_columns,
            subset_duplicates_columns=subset_duplicates_columns,
            expected_val_dtypes=expected_val_dtypes,
            msg_prefix=msg_prefix,
            allowed_nan_value_prop=allowed_nan_value_prop,
            arg_dict=d,
        )

        if df_failures:
            failure_dicts.append(df_failures)

    if not failure_dicts:
        msg.good(
            f"Checked {len(unique_subset_dicts)} predictor_dfs, all returned appropriate dfs",
        )
    else:
        raise ValueError(f"{failure_dicts}")



"""Recommended logger to use."""
import logging
from datetime import datetime

import coloredlogs

from psycop_feature_generation.application_modules.project_setup import ProjectInfo


def init_root_logger(
    project_info: ProjectInfo,
    stdout_log_level: int = logging.INFO,
    log_file_level: int = logging.DEBUG,
) -> None:
    """Initializes the root logger with a file handler and a stream handler."""
    # Get the root logger
    root_log = logging.getLogger()

    # Set the root logger's level to the minimum of the stdout and file log
    # The root logger acts as the trunk of a tree, only messages with a level
    # equal to or higher than the root logger's level will be passed on to
    # the root logger's its branches (its handlers)
    root_log.setLevel(min(stdout_log_level, log_file_level))

    # Install the coloredlogs module on the root logger
    # to get prettier console output and to add a streamhandler,
    # which will write all logging messages from the root logger to
    # stdout
    coloredlogs.install(
        level=stdout_log_level,
        fmt="%(asctime)s [%(levelname)s] %(message)s",
    )

    # Create a timestamped file handler which writes all logging messages from
    # the root logger to a file
    now = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
    log_dir_path = project_info.feature_set_path / "logs"
    log_dir_path.mkdir(exist_ok=True, parents=True)

    file_handler = logging.FileHandler(filename=log_dir_path / f"{now}.log")
    file_handler.setLevel(log_file_level)
    root_log.addHandler(file_handler)



"""Testing loading of coercion functions."""

# pylint: disable=non-ascii-name

import psycop_feature_generation.loaders.raw.load_coercion as c

if __name__ == "__main__":
    df = c.coercion_duration(n_rows=100)
    skema_2 = c.skema_2(n_rows=10000)
    farlighed = c.farlighed(n_rows=20)
    baelte = c.baelte(n_rows=100)



"""Loaders for T2D outcomes."""

# pylint: disable=missing-function-docstring

from typing import Optional

import pandas as pd

from psycop_feature_generation.loaders.raw.sql_load import sql_load
from psycop_feature_generation.utils import data_loaders


@data_loaders.register("t2d")
def t2d(n_rows: Optional[int] = None) -> pd.DataFrame:
    df = sql_load(
        "SELECT dw_ek_borger, timestamp FROM [fct].[psycop_t2d_first_diabetes_t2d] WHERE timestamp IS NOT NULL",
        database="USR_PS_FORSK",
        chunksize=None,
        format_timestamp_cols_to_datetime=True,
        n_rows=n_rows,
    )
    df["value"] = 1

    # 2 duplicates, dropping
    df = df.drop_duplicates(keep="first")

    return df.reset_index(drop=True)


@data_loaders.register("any_diabetes")
def any_diabetes(n_rows: Optional[int] = None):
    df = sql_load(
        "SELECT * FROM [fct].[psycop_t2d_first_diabetes_any] WHERE timestamp IS NOT NULL",
        database="USR_PS_FORSK",
        chunksize=None,
        n_rows=n_rows,
    )

    df = df[["dw_ek_borger", "datotid_first_diabetes_any"]]
    df["value"] = 1

    df.rename(columns={"datotid_first_diabetes_any": "timestamp"}, inplace=True)
    df["timestamp"] = pd.to_datetime(df["timestamp"]).dt.tz_localize(None)

    output = df[["dw_ek_borger", "timestamp", "value"]]
    return output.reset_index(drop=True)



"""Example of how to load IDs from sql."""

from psycop_feature_generation.loaders.raw.sql_load import sql_load

if __name__ == "__main__":
    VIEW = "[psycop_t2d_train]"
    SQL = "SELECT * FROM [fct]." + VIEW
    df = sql_load(SQL, chunksize=None, format_timestamp_cols_to_datetime=False)



"""Loaders for visits to psychiatry."""

import logging
from typing import Literal, Optional

import pandas as pd
from timeseriesflattener.feature_spec_objects import BaseModel

from psycop_feature_generation.loaders.raw.sql_load import sql_load
from psycop_feature_generation.utils import data_loaders

log = logging.getLogger(__name__)


class RawValueSourceSchema(BaseModel):
    """A class for structuring raw values from source schemas.

    Fields:
        view (str): The name of the view for the schema.
        datetime_col (str): The name of the datetime column in the view. This should be defined as the end time of the visit, in order to prevent data leakage.
        value_col (Optional[str]): The name of the column used to calculate the value for the loader. To calculate length of visit, this needs to be the name of the column indicating the starting time of the visit. Defaults to None.
        location_col (Optional[str]): The name of the column indicating which shak unit was responsible for the visit (e.g. 6600). Defaults to None.
        where (str): A where clause that defines additional subsetting for views.
    """

    view: str
    datetime_col: str
    value_col: Optional[str] = None
    location_col: Optional[str] = None
    where_clause: str


def physical_visits(
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = "=",
    where_clause: Optional[str] = None,
    where_separator: Optional[str] = "AND",
    n_rows: Optional[int] = None,
    return_value_as_visit_length_days: Optional[bool] = False,
    visit_types: Optional[
        list[Literal["admissions", "ambulatory_visits", "emergency_visits"]]
    ] = None,
) -> pd.DataFrame:
    """Load pshysical visits to both somatic and psychiatry.

    Args:
        shak_code (Optional[int], optional): SHAK code indicating where to keep/not keep visits from (e.g. 6600). Combines with
            shak_sql_operator, e.g. "!= 6600". Defaults to None, in which case all admissions are kept.
        shak_sql_operator (Optional[str], optional): Operator to use with shak_code. Defaults to "=".
        where_clause (Optional[str], optional): Extra where-clauses to add to the SQL call. E.g. dw_ek_borger = 1. Defaults to None. # noqa: DAR102
        where_separator (Optional[str], optional): Separator between where-clauses. Defaults to "AND".
        n_rows (Optional[int], optional): Number of rows to return. Defaults to None.
        return_value_as_visit_length_days (Optional[bool], optional): Whether to return length of visit in days as the value for the loader. Defaults to False which results in value=1 for all visits.
        visit_types (Optional[list[Literal["admissions", "ambulatory_visits", "emergency_visits"]]], optional): Whether to subset visits by visit types. Defaults to None.

    Returns:
        pd.DataFrame: Dataframe with all physical visits to psychiatry. Has columns dw_ek_borger and timestamp.
    """

    # SHAK = 6600 ≈ in psychiatry

    d = {
        "LPR3": RawValueSourceSchema(
            view="[FOR_LPR3kontakter_psyk_somatik_inkl_2021_feb2022]",
            datetime_col="datotid_lpr3kontaktslut",
            value_col="datotid_lpr3kontaktstart",
            location_col="shakkode_lpr3kontaktansvarlig",
            where_clause="AND [Kontakttype] = 'Fysisk fremmøde'",
        ),
        "ambulatory_visits": RawValueSourceSchema(
            view="[FOR_besoeg_psyk_somatik_LPR2_inkl_2021_feb2022]",
            datetime_col="datotid_slut",
            value_col="datotid_start",
            location_col="shakafskode",
            where_clause="AND ambbesoeg = 1",
        ),
        "emergency_visits": RawValueSourceSchema(
            view="[FOR_akutambulantekontakter_psyk_somatik_LPR2_inkl_2021_feb2022]",
            datetime_col="datotid_slut",
            value_col="datotid_start",
            location_col="afsnit_stam",
            where_clause="",
        ),
        "admissions": RawValueSourceSchema(
            view="[FOR_indlaeggelser_psyk_somatik_LPR2_inkl_2021_feb2022]",
            datetime_col="datotid_udskrivning",
            value_col="datotid_indlaeggelse",
            location_col="shakKode_kontaktansvarlig",
            where_clause="",
        ),
    }

    if visit_types:
        allowed_visit_types = ["admissions", "ambulatory_visits", "emergency_visits"]
        if any(types not in allowed_visit_types for types in visit_types):
            raise ValueError(
                f"Invalid visit type. Allowed types of visits are {allowed_visit_types}.",
            )

        english_to_lpr3_visit_type = {  # pylint: disable=invalid-name
            "admissions": "'Indlæggelse'",
            "ambulatory_visits": "'Ambulant'",
            "emergency_visits": "'Akut ambulant'",
        }
        d = {key: d[key] for key in visit_types + ["LPR3"]}
        english_to_lpr3_visit_type = [
            english_to_lpr3_visit_type[visit] for visit in visit_types
        ]
        d[
            "LPR3"
        ].where_clause += f" AND pt_type IN ({','.join(english_to_lpr3_visit_type)})"

    dfs = []

    for schema in d.values():
        cols = f"{schema.datetime_col}, dw_ek_borger"

        if return_value_as_visit_length_days:
            cols += f", {schema.value_col} AS value_col"

        sql = f"SELECT {cols} FROM [fct].{schema.view} WHERE {schema.datetime_col} IS NOT NULL {schema.where_clause}"

        if shak_code is not None:
            sql += f" AND {schema.location_col} != 'Ukendt'"
            sql += f" AND left({schema.location_col}, {len(str(shak_code))}) {shak_sql_operator} {str(shak_code)}"

        if where_clause is not None:
            sql += f" {where_separator} {where_clause}"

        df = sql_load(sql, database="USR_PS_FORSK", chunksize=None, n_rows=n_rows)
        df.rename(columns={schema.datetime_col: "timestamp"}, inplace=True)

        dfs.append(df)

    # Concat the list of dfs
    output_df = pd.concat(dfs)

    # 0,8% of visits are duplicates. Unsure if overlap between sources or errors in source data. Removing.
    output_df = output_df.drop_duplicates(
        subset=["timestamp", "dw_ek_borger"],
        keep="first",
    )

    # Change value column to length of admission in days
    if return_value_as_visit_length_days:
        output_df["value"] = (
            output_df["timestamp"] - pd.to_datetime(output_df["value_col"])
        ).dt.total_seconds() / 86400
        output_df = output_df.drop(columns="value_col")
    else:
        output_df["value"] = 1

    log.info("Loaded physical visits")

    return output_df.reset_index(drop=True)


@data_loaders.register("physical_visits")
def physical_visits_loader(
    n_rows: Optional[int] = None,
    return_value_as_visit_length_days: Optional[bool] = False,
) -> pd.DataFrame:
    """Load physical visits to all units."""
    return physical_visits(
        n_rows=n_rows,
        return_value_as_visit_length_days=return_value_as_visit_length_days,
    )


@data_loaders.register("physical_visits_to_psychiatry")
def physical_visits_to_psychiatry(
    n_rows: Optional[int] = None,
    timestamps_only: bool = True,
    return_value_as_visit_length_days: Optional[bool] = False,
) -> pd.DataFrame:
    """Load physical visits to psychiatry."""
    df = physical_visits(
        shak_code=6600,
        shak_sql_operator="=",
        n_rows=n_rows,
        return_value_as_visit_length_days=return_value_as_visit_length_days,
    )

    if timestamps_only:
        df = df.drop("value", axis=1)

    return df


@data_loaders.register("physical_visits_to_somatic")
def physical_visits_to_somatic(
    n_rows: Optional[int] = None,
    return_value_as_visit_length_days: Optional[bool] = False,
) -> pd.DataFrame:
    """Load physical visits to somatic."""
    return physical_visits(
        shak_code=6600,
        shak_sql_operator="!=",
        n_rows=n_rows,
        return_value_as_visit_length_days=return_value_as_visit_length_days,
    )


@data_loaders.register("admissions")
def admissions(
    n_rows: Optional[int] = None,
    return_value_as_visit_length_days: Optional[bool] = False,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    """Load admissions."""
    return physical_visits(
        visit_types=["admissions"],
        return_value_as_visit_length_days=return_value_as_visit_length_days,
        n_rows=n_rows,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("ambulatory_visits")
def ambulatory_visits(
    n_rows: Optional[int] = None,
    return_value_as_visit_length_days: Optional[bool] = False,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    """Load ambulatory visits."""
    return physical_visits(
        visit_types=["ambulatory_visits"],
        return_value_as_visit_length_days=return_value_as_visit_length_days,
        n_rows=n_rows,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("emergency_visits")
def emergency_visits(
    n_rows: Optional[int] = None,
    return_value_as_visit_length_days: Optional[bool] = False,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    """Load emergency visits."""
    return physical_visits(
        visit_types=["emergency_visits"],
        return_value_as_visit_length_days=return_value_as_visit_length_days,
        n_rows=n_rows,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )


@data_loaders.register("ambulatory_and_emergency_visits")
def ambulatory_and_emergency_visits(
    n_rows: Optional[int] = None,
    return_value_as_visit_length_days: Optional[bool] = False,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    """Load emergency visits."""
    return physical_visits(
        visit_types=["ambulatory_visits", "emergency_visits"],
        return_value_as_visit_length_days=return_value_as_visit_length_days,
        n_rows=n_rows,
        shak_code=shak_code,
        shak_sql_operator=shak_sql_operator,
    )



"""Utility functions for SQL loading."""

import urllib
import urllib.parse
from collections.abc import Generator
from typing import Optional, Union

import pandas as pd
from sqlalchemy import create_engine


def sql_load(
    query: str,
    server: str = "BI-DPA-PROD",
    database: str = "USR_PS_Forsk",
    chunksize: Optional[int] = None,
    format_timestamp_cols_to_datetime: Optional[bool] = True,
    n_rows: Optional[int] = None,
) -> Union[pd.DataFrame, Generator[pd.DataFrame, None, None]]:
    """Function to load a SQL query. If chunksize is None, all data will be
    loaded into memory. Otherwise, will stream the data in chunks of chunksize
    as a generator.

    Args:
        query (str): The SQL query
        server (str): The BI server
        database (str): The BI database
        chunksize (int, optional): Defaults to 1000.
        format_timestamp_cols_to_datetime (bool, optional): Whether to format all
            columns with "datotid" in their name as pandas datetime. Defaults to true.
        n_rows (int, optional): Defaults to None. If specified, only returns the first n rows.

    Returns:
        Union[pd.DataFrame, Generator[pd.DataFrame]]: DataFrame or generator of DataFrames

    Example:
        # From USR_PS_Forsk
        >>> view = "[FOR_SFI_fritekst_resultat_udfoert_i_psykiatrien_aendret_2011]"
        >>> sql = "SELECT * FROM [fct]." + view
        >>> df = sql_load(sql, chunksize = None)
    """
    driver = "SQL Server"
    params = urllib.parse.quote(
        f"DRIVER={driver};SERVER={server};DATABASE={database};Trusted_Connection=yes",
    )

    if n_rows:
        query = query.replace("SELECT", f"SELECT TOP {n_rows} ")

    engine = create_engine(f"mssql+pyodbc:///?odbc_connect={params}")

    conn = engine.connect().execution_options(
        stream_results=True,
        fast_executemany=True,
    )

    df = pd.read_sql(query, conn, chunksize=chunksize)

    if format_timestamp_cols_to_datetime:
        datetime_col_names = [
            colname
            for colname in df.columns
            if any(substr in colname.lower() for substr in ["datotid", "timestamp"])
        ]

        df[datetime_col_names] = df[datetime_col_names].apply(pd.to_datetime)

    conn.close()
    engine.dispose()

    return df



"""Example of."""

import logging
from typing import Optional, Union

import pandas as pd

from psycop_feature_generation.loaders.raw.sql_load import sql_load

log = logging.getLogger(__name__)


def str_to_sql_match_logic(
    code_to_match: str,
    code_sql_col_name: str,
    load_diagnoses: bool,
    match_with_wildcard: bool,
) -> str:
    """Generate SQL match logic from a single string.

    Args:
        code_to_match (list[str]): List of strings to match.
        code_sql_col_name (str): Name of the SQL column containing the codes.
        load_diagnoses (bool): Whether to load diagnoses or medications. Determines the logic. See calling function for more.
        match_with_wildcard (bool): Whether to match on icd_code* / atc_code* or only icd_code / atc_code.
    """
    base_query = f"lower({code_sql_col_name}) LIKE '%{code_to_match.lower()}"

    if match_with_wildcard:
        return f"{base_query}%'"

    if load_diagnoses:
        return f"{base_query}' OR {base_query}#%'"

    return f"{base_query}'"


def list_to_sql_logic(
    codes_to_match: list[str],
    code_sql_col_name: str,
    load_diagnoses: bool,
    match_with_wildcard: bool,
):
    """Generate SQL match logic from a list of strings.

    Args:
        codes_to_match (list[str]): List of strings to match.
        code_sql_col_name (str): Name of the SQL column containing the codes.
        load_diagnoses (bool): Whether to load diagnoses or medications. Determines the logic. See calling function for more.
        match_with_wildcard (bool): Whether to match on icd_code* / atc_code* or only icd_code / atc_code.
    """
    match_col_sql_strings = []

    for code_str in codes_to_match:
        base_query = f"lower({code_sql_col_name}) LIKE '%{code_str.lower()}"

        if match_with_wildcard:
            match_col_sql_strings.append(
                f"{base_query}%'",
            )
        else:
            # If the string is at the end of diagnosegruppestreng, it doesn't end with a hashtag
            match_col_sql_strings.append(f"{base_query}'")

            if load_diagnoses:
                # If the string is at the beginning of diagnosegruppestreng, it doesn't start with a hashtag
                match_col_sql_strings.append(
                    f"lower({code_sql_col_name}) LIKE '{code_str.lower()}#%'",
                )

    return " OR ".join(match_col_sql_strings)


def load_from_codes(
    codes_to_match: Union[list[str], str],
    load_diagnoses: bool,
    code_col_name: str,
    source_timestamp_col_name: str,
    view: str,
    output_col_name: Optional[str] = None,
    match_with_wildcard: bool = True,
    n_rows: Optional[int] = None,
    exclude_codes: Optional[list[str]] = None,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
    shak_location_col: Optional[str] = None,
    shak_code: Optional[int] = None,
    shak_sql_operator: Optional[str] = None,
) -> pd.DataFrame:
    """Load the visits that have diagnoses that match icd_code or atc code from
    the beginning of their adiagnosekode or atc code string. Aggregates all
    that match.

    Args:
        codes_to_match (Union[list[str], str]): Substring(s) to match diagnoses or medications for.
            Diagnoses: Matches any diagnoses, whether a-diagnosis, b-diagnosis.
            Both: If a list is passed, will count as a match if any of the icd_codes or at codes in the list match.
        load_diagnoses (bool): Determines which mathing logic is employed. If True, will load diagnoses. If False, will load medications.
            Diagnoses must be able to split a string like this:
                A:DF431#+:ALFC3#B:DF329
            Which means that if match_with_wildcard is False, we must match on *icd_code# or *icd_code followed by nothing. If it's true, we can match on *icd_code*.
        code_col_name (str): Name of column containing either diagnosis (icd) or medication (atc) codes.
            Takes either 'diagnosegruppestreng' or 'atc' as input.
        source_timestamp_col_name (str): Name of the timestamp column in the SQL
            view.
        view (str): Name of the SQL view to load from.
        output_col_name (str, optional): Name of new column string. Defaults to
            None.
        match_with_wildcard (bool, optional): Whether to match on icd_code* / atc_code*.
            Defaults to true.
        n_rows: Number of rows to return. Defaults to None.
        exclude_codes (list[str], optional): Drop rows if their code is in this list. Defaults to None.
        administration_route (str, optional): Whether to subset by a specific administration route, e.g. 'OR', 'IM' or 'IV'. Defaults to None.
        administration_method (str, optional): Whether to subset by method of administration, e.g. 'PN' or 'Fast'. Defaults to None.
        shak_location_col (str, optional): Name of column containing shak code. Defaults to None. Combine with shak_code and shak_sql_operator.
        shak_code (int, optional): Shak code indicating where to keep/not keep visits from (e.g. 6600). Defaults to None.
        shak_sql_operator (str, optional): Operator indicating how to filter shak_code, e.g. "!= 6600" or "= 6600". Defaults to None.

    Returns:
        pd.DataFrame: A pandas dataframe with dw_ek_borger, timestamp and
            output_col_name = 1
    """
    fct = f"[{view}]"

    if isinstance(codes_to_match, list) and len(codes_to_match) > 1:
        match_col_sql_str = list_to_sql_logic(
            codes_to_match=codes_to_match,
            code_sql_col_name=code_col_name,
            load_diagnoses=load_diagnoses,
            match_with_wildcard=match_with_wildcard,
        )
    elif isinstance(codes_to_match, str):
        match_col_sql_str = str_to_sql_match_logic(
            code_to_match=codes_to_match,
            code_sql_col_name=code_col_name,
            load_diagnoses=load_diagnoses,
            match_with_wildcard=match_with_wildcard,
        )
    else:
        raise ValueError("codes_to_match must be either a list or a string.")

    sql = (
        f"SELECT dw_ek_borger, {source_timestamp_col_name}, {code_col_name} "
        + f"FROM [fct].{fct} WHERE {source_timestamp_col_name} IS NOT NULL AND ({match_col_sql_str})"
    )

    if shak_code is not None:
        sql += f" AND left({shak_location_col}, {len(str(shak_code))}) {shak_sql_operator} {str(shak_code)}"

    if administration_method:
        allowed_administration_methods = (
            "Fast",
            "PN",
            "Engangs",
            "Alternerende",
            "Kontinuerlig",
            "Skema2",
            "Skema",
            "Tidspunkter",
        )
        if administration_method not in allowed_administration_methods:
            log.warning(
                "Value for administration method does not exist, returning 0 rows. "
                + f"Allowed values are {allowed_administration_methods}.",
            )
        sql += f" AND type_kodetekst = '{administration_method}'"

    if administration_route:
        allowed_administration_routes = (
            "OR",
            "IV",
            "IH",
            "IM",
            "SC",
            "KU",
            "IA",
            "IR",
            "PR",
            "IN",
            "OK",
            "TD",
            "PO",
            "SL",
            "BS",
            "DE",
            "ED",
            "CO",
            "VA",
            "LO",
            "PE",
            "AU",
            "RE",
            "IE",
            "IU",
            "UR",
            "GA",
            "OG",
            "OS",
            "IC",
            "OM",
            "ET",
            "HE",
            "IB",
            "BR",
            "KO",
            "VI",
            "EC",
            "IL",
            "IP",
            "IT",
            "MP",
            "CA",
            "IO",
            "IS",
            "CE",
            "ID",
            "ES",
            "SM",
            "TR",
            "PA",
            "PT",
            "TO",
            "PD",
            "ON",
            "BU",
            "GI",
            "OF",
            "AM",
            "CB",
            "EL",
            "PV",
            "LY",
            "XA",
            "IF",
            "AL",
            "DI",
            "PN",
            "PC",
            "BD",
            "IG",
            "HS",
            "TU",
            "PJ",
            "LF",
        )
        if administration_route not in allowed_administration_routes:
            log.warning(
                "Value for administration route does not exist, returning 0 rows. "
                + f"Allowed values are {allowed_administration_routes}.",
            )
        sql += f" AND admvej_kodetekst = '{administration_route}'"

    df = sql_load(sql, database="USR_PS_FORSK", chunksize=None, n_rows=n_rows)

    if exclude_codes:
        # Drop all rows whose code_col_name is in exclude_codes
        df = df[~df[code_col_name].isin(exclude_codes)]

    if output_col_name is None:
        if isinstance(codes_to_match, list):
            output_col_name = "_".join(codes_to_match)
        else:
            output_col_name = codes_to_match

    df[output_col_name] = 1

    df.drop([f"{code_col_name}"], axis="columns", inplace=True)

    return df.rename(
        columns={
            source_timestamp_col_name: "timestamp",
        },
    )



"""Load text data from sql warehouse."""

from functools import partial
from multiprocessing import Pool
from typing import Iterable, Optional, Union

import pandas as pd

from psycop_feature_generation.loaders.raw.sql_load import sql_load
from psycop_feature_generation.utils import data_loaders


def get_all_valid_note_types() -> set[str]:
    """Returns a set of valid note types. Notice that 'Konklusion' is replaced
    by 'Vurdering/konklusion' in 2020, so make sure to use both. 'Ordination'
    was replaced by 'Ordination, Psykiatry' in 2022, but 'Ordination,
    Psykiatri' is not included in the table. Use with caution.

    Returns:
        Set[str]: Set of valid note types
    """
    return {
        "Observation af patient, Psykiatri",
        "Samtale med behandlingssigte",
        "Ordination",  # OBS replaced "Ordination, Psykiatri" in 01/02-22
        # but is not included in this table. Use with caution
        "Aktuelt psykisk",
        "Aktuelt socialt, Psykiatri",
        "Aftaler, Psykiatri",
        "Medicin",
        "Aktuelt somatisk, Psykiatri",
        "Objektivt psykisk",
        "Kontaktårsag",
        "Telefonkonsultation",
        "Journalnotat",
        "Telefonnotat",
        "Objektivt, somatisk",
        "Plan",
        "Semistruktureret diagnostisk interview",
        "Vurdering/konklusion",
    }


def _load_notes_for_year(
    note_types: Union[str, list[str]],
    year: str,
    view: Optional[str] = "FOR_SFI_fritekst_resultat_udfoert_i_psykiatrien_aendret",
    n_rows: Optional[int] = None,
) -> pd.DataFrame:
    """Loads clinical notes from sql from a specified year and matching
    specified note types.

    Args:
        note_names (Union[str, list[str]]): Which types of notes to load.
        year (str): Which year to load
        view (str, optional): Which table to load.
            Defaults to "[FOR_SFI_fritekst_resultat_udfoert_i_psykiatrien_aendret".
        n_rows (Optional[int], optional): Number of rows to load. Defaults to None.

    Returns:
        pd.DataFrame: Dataframe with clinical notes
    """

    sql = (
        "SELECT dw_ek_borger, datotid_senest_aendret_i_sfien, fritekst"
        + f" FROM [fct].[{view}_{year}_inkl_2021_feb2022]"
        + f" WHERE overskrift IN {note_types}"
    )
    return sql_load(
        sql,
        database="USR_PS_FORSK",
        chunksize=None,
        n_rows=n_rows,
    )


def load_notes(
    note_types: Union[str, Iterable[str]],
    n_rows: Optional[int] = None,
) -> pd.DataFrame:
    """Loads all clinical notes that match the specified note from all years.

    Args:
        note_types (Union[str, list[str]]): Which note types to load. See
            `get_all_valid_note_types()` for valid note types.
        n_rows (Optional[int], optional): How many rows to load. Defaults to None.

    Raises:
        ValueError: If given invalid note type

    Returns:
        pd.DataFrame: Featurized clinical notes
    """
    if isinstance(note_types, str):
        note_types = list(note_types)
    # check for invalid note types
    if not set(note_types).issubset(get_all_valid_note_types()):
        raise ValueError(
            "Invalid note type. Valid note types are: "
            + str(get_all_valid_note_types()),
        )

    # convert note_types to sql query
    note_types = "('" + "', '".join(note_types) + "')"

    view = "FOR_SFI_fritekst_resultat_udfoert_i_psykiatrien_aendret"

    load_and_featurize = partial(
        _load_notes_for_year,
        note_types=note_types,
        view=view,
        n_rows=n_rows,
    )

    years = list(range(2011, 2022))

    with Pool(processes=len(years)) as p:
        dfs = p.map(load_and_featurize, [str(y) for y in years])

    df = pd.concat(dfs)
    df = df.rename(
        {"datotid_senest_aendret_i_sfien": "timestamp", "fritekst": "text"},
        axis=1,
    )
    return df


@data_loaders.register("all_notes")
def load_all_notes(
    n_rows: Optional[int] = None,
) -> pd.DataFrame:
    """Returns all notes from all years.

    Args:
        n_rows (Optional[int], optional): Number of rows to load. Defaults to None.

    Returns:
        pd.DataFrame: (Featurized) notes
    """
    return load_notes(
        note_types=get_all_valid_note_types(),
        n_rows=n_rows,
    )


@data_loaders.register("aktuelt_psykisk")
def load_aktuel_psykisk(
    n_rows: Optional[int] = None,
) -> pd.DataFrame:
    """Returns 'Aktuelt psykisk' notes from all years.

    Args:
        n_rows (Optional[int], optional): Number of rows to load. Defaults to None.

    Returns:
        pd.DataFrame: (Featurized) notes
    """
    return load_notes(
        note_types="Aktuelt psykisk",
        n_rows=n_rows,
    )


@data_loaders.register("load_note_types")
def load_arbitrary_notes(
    note_types: Union[str, list[str]],
    n_rows: Optional[int] = None,
) -> pd.DataFrame:
    """Returns one or multiple note types from all years.

    Args:
        note_types (Union[str, list[str]]): Which note types to load. See
            `get_all_valid_note_types()` for a list of valid note types.
        n_rows (Optional[int], optional): Number of rows to load. Defaults to None.

    Returns:
        pd.DataFrame: (Featurized) notes
    """
    return load_notes(
        note_types,
        n_rows=n_rows,
    )



"""Train a TF-IDF featurizer on train set of all clinical notes."""
# import dill as pkl
# import pandas as pd
# from sklearn.feature_extraction.text import TfidfVectorizer
# from wasabi import Printer

# from psycop_feature_generation.loaders.raw.load_ids import load_ids
# from psycop_feature_generation.loaders.raw.load_text import load_all_notes
# from psycop_feature_generation.utils import FEATURIZERS_PATH, RELATIVE_PROJECT_ROOT


# def create_tfidf_vectorizer(
#     ngram_range: tuple[int, int] = (1, 2),
#     max_df: float = 0.95,
#     min_df: float = 0.01,
#     max_features: int = 100,
# ) -> TfidfVectorizer:
#     """Creates a TF-IDF vectorizer with a whitespace tokenizer.

#     Args:
#         ngram_range (Tuple[int, int], optional): How many ngrams to make. Defaults to (1, 2).
#         max_df (float, optional): Removes words occuring in max_df proportion of documents.
#             Defaults to 0.95.
#         min_df (float, optional): Removes words occuring in less than min_df proportion
#             of documents. Defaults to 0.01.
#         max_features (int, optional): How many features to create. Defaults to 100.

#     Returns:
#         TfidfVectorizer: Sklearn TF-IDF vectorizer
#     """
#     return TfidfVectorizer(
#         ngram_range=ngram_range,
#         tokenizer=lambda x: x.split(" "),
#         lowercase=True,
#         max_df=max_df,  # remove very common words
#         min_df=min_df,  # remove very rare words
#         max_features=max_features,
#     )


# if __name__ == "__main__":

#     SYNTHETIC = False
#     OVERTACI = True

#     msg = Printer(timestamp=True)

#     # Train TF-IDF model on all notes
#     if OVERTACI:
#         if not FEATURIZERS_PATH.exists():
#             FEATURIZERS_PATH.mkdir()

#         text = load_all_notes(  # pylint: disable=undefined-variable
#             featurizer=None,
#             n_rows=None,
#             featurizer_kwargs=None,
#         )

#         # Subset only train set
#         train_ids = load_ids(split="train")
#         train_ids = train_ids["dw_ek_borger"].unique()
#         text = text[text["dw_ek_borger"].isin(train_ids)]
#         text = text["text"].tolist()

#         for n_features in [100, 500, 1000]:
#             msg.info(f"Fitting tf-idf with {n_features} features..")
#             vectorizer = create_tfidf_vectorizer(max_features=n_features)
#             vectorizer.fit(text)

#             with open(FEATURIZERS_PATH / f"tfidf_{n_features}.pkl", "wb") as f:
#                 pkl.dump(vectorizer, f)

#             vocab = ["tfidf-" + word for word in vectorizer.get_feature_names()]
#             with open(  # type: ignore # pylint: disable=unspecified-encoding
#                 FEATURIZERS_PATH / f"tfidf_{n_features}_vocab.txt",
#                 "w",
#             ) as f:
#                 f.write("\n".join(vocab))  # type: ignore

#     # train TF-IDF on synthetic data
#     if SYNTHETIC:
#         test_path = RELATIVE_PROJECT_ROOT / "tests" / "test_data"
#         save_dir = test_path / "test_tfidf"
#         if not save_dir.exists():
#             save_dir.mkdir()

#         text = pd.read_csv(test_path / "synth_txt_data.csv")
#         text = text["text"].dropna().tolist()

#         msg.info("Fitting tf-idf with 10 features..")
#         vectorizer = create_tfidf_vectorizer(max_features=10)
#         vectorizer.fit(text)

#         with open(save_dir / "tfidf_10.pkl", "wb") as f:
#             pkl.dump(vectorizer, f)



"""Example of how to get tfidf vocab."""

# pylint: disable=missing-function-docstring

from psycop_feature_generation.featurizers import FEATURIZERS_PATH


def get_tfidf_vocab(
    n_features: int,
) -> list[str]:  # pylint: disable=missing-function-docstring
    with open(  # pylint: disable=unspecified-encoding
        FEATURIZERS_PATH / f"tfidf_{str(n_features)}.txt",
        "r",
    ) as f:
        return f.read().splitlines()


TFIDF_VOCAB = {n: get_tfidf_vocab(n) for n in [100, 500, 1000]}



"""Example loader for physical visits."""
import psycop_feature_generation.loaders.raw as r

if __name__ == "__main__":
    df = r.load_visits.physical_visits(n_rows=1000)
    psych = r.load_visits.physical_visits_to_psychiatry(n_rows=1000)
    somatic = r.load_visits.physical_visits_to_somatic(n_rows=1000)



"""Utilities for saving a dataset to disk."""
import logging
from typing import Literal

import numpy as np
import pandas as pd

import psycop_feature_generation.loaders
from psycop_feature_generation.application_modules.project_setup import ProjectInfo
from psycop_feature_generation.application_modules.wandb_utils import (
    wandb_alert_on_exception,
)
from psycop_feature_generation.utils import write_df_to_file

log = logging.getLogger(__name__)


def save_split_to_disk(
    project_info: ProjectInfo,
    split_df: pd.DataFrame,
    split_name: str,
):
    """Save split to disk."""
    # Version table with current date and time
    filename = (
        f"{project_info.feature_set_prefix}_{split_name}.{project_info.dataset_format}"
    )
    log.info(f"Saving {filename} to disk")

    file_path = project_info.feature_set_path / filename

    write_df_to_file(df=split_df, file_path=file_path)

    log.info(f"{split_name}: Succesfully saved to {file_path}")


def filter_by_split_ids(
    flattened_df: pd.DataFrame,
    split_id_df: pd.DataFrame,
    split_name: str,
):
    """Filter dataframe by split ids."""
    # Find IDs which are in split_ids, but not in flattened_df
    flattened_df_ids = flattened_df["dw_ek_borger"].unique()
    split_ids: pd.Series = split_id_df["dw_ek_borger"].unique()

    ids_in_split_but_not_in_flattened_df = split_ids[
        ~np.isin(split_ids, flattened_df_ids)
    ]

    log.warning(
        f"{split_name}: There are {len(ids_in_split_but_not_in_flattened_df)} ({round(len(ids_in_split_but_not_in_flattened_df) / len(split_ids) * 100, 2)}%) ids which are in {split_name}_ids but not in flattened_df_ids, will get dropped during merge. If examining patients based on physical visits, see 'OBS: Patients without physical visits' on the wiki for more info.",
    )

    split_df = pd.merge(flattened_df, split_id_df, how="inner", validate="m:1")
    return split_df


def get_split_id_df(split_name: Literal["train", "val", "test"]) -> pd.DataFrame:
    """Get a dataframe with the splits ids."""
    split_id_df = psycop_feature_generation.loaders.raw.load_ids(
        split=split_name,
    )

    return split_id_df


@wandb_alert_on_exception
def split_and_save_dataset_to_disk(
    flattened_df: pd.DataFrame,
    project_info: ProjectInfo,
):
    """Split and save to disk.

    Args:
        flattened_df (pd.DataFrame): Flattened dataframe.
        project_info (ProjectInfo): Project info.
    """
    for split_name in ("train", "val", "test"):
        split_id_df = get_split_id_df(split_name=split_name)  # type: ignore

        split_df = filter_by_split_ids(
            flattened_df=flattened_df,
            split_name=split_name,
            split_id_df=split_id_df,
        )

        save_split_to_disk(
            project_info=project_info,
            split_df=split_df,
            split_name=split_name,
        )



"""Loaders for lab results loading."""

from typing import Optional, Union

import pandas as pd

from psycop_feature_generation.loaders.non_numerical_coercer import (
    multiply_inequalities_in_df,
)
from psycop_feature_generation.loaders.raw.sql_load import sql_load
from psycop_feature_generation.utils import data_loaders

# pylint: disable=missing-function-docstring


def load_non_numerical_values_and_coerce_inequalities(
    blood_sample_id: Union[str, list[str]],
    n_rows: Optional[int],
    view: str,
    ineq2mult: Optional[dict[str, float]] = None,
) -> pd.DataFrame:
    """Load non-numerical values for a blood sample.

    Args:
        blood_sample_id (Union[str, list]): The blood_sample_id, typically an NPU code. If a list, concatenates the values. # noqa: DAR102
        n_rows (Optional[int]): Number of rows to return. Defaults to None.
        view (str): The view to load from.
        ineq2mult (dict[str, float]): A dictionary mapping inequalities to a multiplier. Defaults to None.

    Returns:
        pd.DataFrame: A dataframe with the non-numerical values.
    """
    cols = "dw_ek_borger, datotid_sidstesvar, svar"

    if isinstance(blood_sample_id, list):
        npu_codes = ", ".join(
            [f"'{x}'" for x in blood_sample_id],  # pylint: disable=not-an-iterable
        )

        npu_where = f"npukode in ({npu_codes})"
    else:
        npu_where = f"npukode = '{blood_sample_id}'"

    sql = f"SELECT {cols} FROM [fct].{view} WHERE datotid_sidstesvar IS NOT NULL AND {npu_where} AND numerisksvar IS NULL AND (left(Svar,1) = '>' OR left(Svar, 1) = '<')"

    df = sql_load(
        sql,
        database="USR_PS_FORSK",
        chunksize=None,
        n_rows=n_rows,
    )

    df.rename(
        columns={"datotid_sidstesvar": "timestamp", "svar": "value"},
        inplace=True,
    )

    return multiply_inequalities_in_df(df, ineq2mult=ineq2mult)


def load_numerical_values(
    blood_sample_id: str,
    n_rows: Optional[int],
    view: str,
) -> pd.DataFrame:
    """Load numerical values for a blood sample.

    Args:
        blood_sample_id (str): The blood_sample_id, typically an NPU code.  # noqa: DAR102
        n_rows (Optional[int]): Number of rows to return. Defaults to None.
        view (str): The view to load from.

    Returns:
        pd.DataFrame: A dataframe with the numerical values.
    """

    cols = "dw_ek_borger, datotid_sidstesvar, numerisksvar"

    if isinstance(blood_sample_id, list):
        npu_codes = ", ".join(
            [f"'{x}'" for x in blood_sample_id],  # pylint: disable=not-an-iterable
        )

        npu_where = f"npukode in ({npu_codes})"
    else:
        npu_where = f"npukode = '{blood_sample_id}'"

    sql = f"SELECT {cols} FROM [fct].{view} WHERE datotid_sidstesvar IS NOT NULL AND {npu_where} AND numerisksvar IS NOT NULL"
    df = sql_load(
        sql,
        database="USR_PS_FORSK",
        chunksize=None,
        n_rows=n_rows,
    )

    df.rename(
        columns={"datotid_sidstesvar": "timestamp", "numerisksvar": "value"},
        inplace=True,
    )

    return df


def load_cancelled(
    blood_sample_id: str,
    n_rows: Optional[int],
    view: str,
) -> pd.DataFrame:
    """Load cancelled samples for a blood sample.

    Args:
        blood_sample_id (str): The blood_sample_id, typically an NPU code.  # noqa: DAR102
        n_rows (Optional[int]): Number of rows to return. Defaults to None.
        view (str): The view to load from.

    Returns:
        pd.DataFrame: A dataframe with the timestamps for cancelled values.
    """
    cols = "dw_ek_borger, datotid_sidstesvar"

    if isinstance(blood_sample_id, list):
        npu_codes = ", ".join(
            [f"'{x}'" for x in blood_sample_id],  # pylint: disable=not-an-iterable
        )

        npu_where = f"npukode in ({npu_codes})"
    else:
        npu_where = f"npukode = '{blood_sample_id}'"

    sql = f"SELECT {cols} FROM [fct].{view} {npu_where} AND datotid_sidstesvar IS NOT NULL AND Svar == 'Aflyst' AND (left(Svar,1) == '>' OR left(Svar, 1) == '<')"

    df = sql_load(
        sql,
        database="USR_PS_FORSK",
        chunksize=None,
        n_rows=n_rows,
    )

    # Create the value column == 1, since all timestamps here are from cancelled blood samples
    df["value"] = 1

    df.rename(
        columns={"datotid_sidstesvar": "timestamp"},
        inplace=True,
    )

    return df


def load_all_values(
    blood_sample_id: str,
    n_rows: Optional[int],
    view: str,
) -> pd.DataFrame:
    """Load all samples for a blood sample.

    Args:
        blood_sample_id (str): The blood_sample_id, typically an NPU code.  # noqa: DAR102
        n_rows (Optional[int]): Number of rows to return. Defaults to None.
        view (str): The view to load from.

    Returns:
        pd.DataFrame: A dataframe with all values.
    """
    cols = "dw_ek_borger, datotid_sidstesvar, svar"

    if isinstance(blood_sample_id, list):
        npu_codes = ", ".join(
            [f"'{x}'" for x in blood_sample_id],  # pylint: disable=not-an-iterable
        )

        npu_where = f"npukode in ({npu_codes})"
    else:
        npu_where = f"npukode = '{blood_sample_id}'"

    sql = f"SELECT {cols} FROM [fct].{view} WHERE datotid_sidstesvar IS NOT NULL AND {npu_where}"

    df = sql_load(
        sql,
        database="USR_PS_FORSK",
        chunksize=None,
        n_rows=n_rows,
    )

    df.rename(
        columns={"datotid_sidstesvar": "timestamp", "svar": "value"},
        inplace=True,
    )

    return df


def blood_sample(
    blood_sample_id: Union[str, list],
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    """Load a blood sample.

    Args:
        blood_sample_id (Union[str, list]): The blood_sample_id, typically an NPU code. If a list, concatenates the values. # noqa: DAR102
        n_rows: Number of rows to return. Defaults to None.
        values_to_load (str): Which values to load. Takes either "numerical", "numerical_and_coerce", "cancelled" or "all". Defaults to None, which is coerced to "all".

    Returns:
        pd.DataFrame
    """
    view = "[FOR_labka_alle_blodprover_inkl_2021_feb2022]"

    allowed_values_to_load = [
        "numerical",
        "numerical_and_coerce",
        "cancelled",
        "all",
        None,
    ]

    dfs = []

    if values_to_load not in allowed_values_to_load:
        raise ValueError(
            f"values_to_load must be one of {allowed_values_to_load}, not {values_to_load}",
        )

    if values_to_load is None:
        values_to_load = "all"

    fn_dict = {
        "coerce": load_non_numerical_values_and_coerce_inequalities,
        "numerical": load_numerical_values,
        "cancelled": load_cancelled,
        "all": load_all_values,
    }

    sources_to_load = [k for k in fn_dict if k in values_to_load]

    if n_rows:
        n_rows_per_fn = int(n_rows / len(sources_to_load))
    else:
        n_rows_per_fn = None

    for k in sources_to_load:  # pylint: disable=invalid-name
        dfs.append(
            fn_dict[k](  # type: ignore
                blood_sample_id=blood_sample_id,
                n_rows=n_rows_per_fn,
                view=view,
            ),
        )

    # Concatenate dfs
    if len(dfs) > 1:
        df = pd.concat(dfs)
    else:
        df = dfs[0]

    return df.reset_index(drop=True).drop_duplicates(
        subset=["dw_ek_borger", "timestamp", "value"],
        keep="first",
    )


@data_loaders.register("hba1c")
def hba1c(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU27300",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("scheduled_glc")
def scheduled_glc(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    npu_suffixes = [
        "08550",
        "08551",
        "08552",
        "08553",
        "08554",
        "08555",
        "08556",
        "08557",
        "08558",
        "08559",
        "08560",
        "08561",
        "08562",
        "08563",
        "08564",
        "08565",
        "08566",
        "08567",
        "08893",
        "08894",
        "08895",
        "08896",
        "08897",
        "08898",
        "08899",
        "08900",
        "08901",
        "08902",
        "08903",
        "08904",
        "08905",
        "08906",
        "08907",
        "08908",
        "08909",
        "08910",
        "08911",
        "08912",
        "08913",
        "08914",
        "08915",
        "08916",
    ]

    blood_sample_ids = [f"NPU{suffix}" for suffix in npu_suffixes]

    return blood_sample(
        blood_sample_id=blood_sample_ids,
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("unscheduled_p_glc")
def unscheduled_p_glc(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    npu_suffixes = [
        "02192",
        "21533",
        "21531",
    ]

    dnk_suffixes = ["35842"]

    blood_sample_ids = [f"NPU{suffix}" for suffix in npu_suffixes]
    blood_sample_ids += [f"DNK{suffix}" for suffix in dnk_suffixes]

    return blood_sample(
        blood_sample_id=blood_sample_ids,
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("triglycerides")
def triglycerides(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU04094",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("fasting_triglycerides")
def fasting_triglycerides(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU03620",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("hdl")
def hdl(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU01567",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("ldl")
def ldl(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id=["NPU01568", "AAB00101"],
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("fasting_ldl")
def fasting_ldl(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id=["NPU10171", "AAB00102"],
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("alat")
def alat(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU19651",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("asat")
def asat(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU19654",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("lymphocytes")
def lymphocytes(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU02636",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("leukocytes")
def leukocytes(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU02593",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("crp")
def crp(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU19748",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("creatinine")
def creatinine(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id=["NPU18016", "ASS00355", "ASS00354"],
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("egfr")
def egfr(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id=["DNK35302", "DNK35131", "AAB00345", "AAB00343"],
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("albumine_creatinine_ratio")
def albumine_creatinine_ratio(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU19661",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("cyp21a2")
def cyp21a2(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU19053",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("cyp2c19")
def cyp2c19(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU19309",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("cyp2c9")
def cyp2c9(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU32095",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("cyp3a5")
def cyp3a5(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU27992",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("cyp2d6")
def cyp2d6(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU19308",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("p_lithium")
def p_lithium(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU02613",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("p_clozapine")
def p_clozapine(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU04114",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("p_olanzapine")
def p_olanzapine(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU09358",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("p_aripiprazol")
def p_aripiprazol(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU26669",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("p_risperidone")
def p_risperidone(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU04868",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("p_paliperidone")
def p_paliperidone(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU18359",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("p_haloperidol")
def p_haloperidol(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU03937",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("p_amitriptyline")
def p_amitriptyline(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU01224",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("p_nortriptyline")
def p_nortriptyline(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU02923",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("p_clomipramine")
def p_clomipramine(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU01616",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("p_paracetamol")
def p_paracetamol(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU03024",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("p_ibuprofen")
def p_ibuprofen(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU08794",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )


@data_loaders.register("p_ethanol")
def p_ethanol(
    n_rows: Optional[int] = None,
    values_to_load: str = "numerical_and_coerce",
) -> pd.DataFrame:
    return blood_sample(
        blood_sample_id="NPU01992",
        n_rows=n_rows,
        values_to_load=values_to_load,
    )



"""Flatten the dataset."""
import logging
from typing import Optional

import pandas as pd
import psutil
from timeseriesflattener.feature_cache.cache_to_disk import DiskCache
from timeseriesflattener.feature_spec_objects import _AnySpec
from timeseriesflattener.flattened_dataset import TimeseriesFlattener

from psycop_feature_generation.application_modules.filter_prediction_times import (
    PredictionTimeFilterer,
)
from psycop_feature_generation.application_modules.project_setup import ProjectInfo
from psycop_feature_generation.application_modules.wandb_utils import (
    wandb_alert_on_exception,
)
from psycop_feature_generation.loaders.raw.load_demographic import birthdays

log = logging.getLogger(__name__)


def filter_prediction_times(
    prediction_times_df: pd.DataFrame,
    project_info: ProjectInfo,
    quarantine_df: Optional[pd.DataFrame] = None,
    quarantine_days: Optional[int] = None,
) -> pd.DataFrame:
    """Filter prediction times.

    Args:
        prediction_times_df (pd.DataFrame): Prediction times dataframe.
            Should contain entity_id and timestamp columns with col_names matching those in project_info.col_names.
        project_info (ProjectInfo): Project info.
        quarantine_df (pd.DataFrame, optional): Quarantine dataframe with "timestamp" and "project_info.col_names.id" columns.
        quarantine_days (int, optional): Number of days to quarantine.

    Returns:
        pd.DataFrame: Filtered prediction times dataframe.
    """
    log.info("Filtering prediction times...")

    filterer = PredictionTimeFilterer(
        prediction_times_df=prediction_times_df,
        entity_id_col_name=project_info.col_names.id,
        quarantine_timestamps_df=quarantine_df,
        quarantine_interval_days=quarantine_days,
    )

    return filterer.filter()


@wandb_alert_on_exception
def create_flattened_dataset(
    feature_specs: list[_AnySpec],
    prediction_times_df: pd.DataFrame,
    drop_pred_times_with_insufficient_look_distance: bool,
    project_info: ProjectInfo,
    quarantine_df: Optional[pd.DataFrame] = None,
    quarantine_days: Optional[int] = None,
) -> pd.DataFrame:
    """Create flattened dataset.

    Args:
        feature_specs (list[_AnySpec]): List of feature specifications of any type.
        project_info (ProjectInfo): Project info.
        prediction_times_df (pd.DataFrame): Prediction times dataframe.
            Should contain entity_id and timestamp columns with col_names matching those in project_info.col_names.
        drop_pred_times_with_insufficient_look_distance (bool): Whether to drop prediction times with insufficient look distance.
            See timeseriesflattener tutorial for more info.
        quarantine_df (pd.DataFrame, optional): Quarantine dataframe with "timestamp" and "project_info.col_names.id" columns.
        quarantine_days (int, optional): Number of days to quarantine. Any prediction time within quarantine_days after the timestamps in quarantine_df will be dropped.

    Returns:
        FlattenedDataset: Flattened dataset.
    """
    filtered_prediction_times_df = filter_prediction_times(
        prediction_times_df=prediction_times_df,
        project_info=project_info,
        quarantine_df=quarantine_df,
        quarantine_days=quarantine_days,
    )

    flattened_dataset = TimeseriesFlattener(
        prediction_times_df=filtered_prediction_times_df,
        n_workers=min(
            len(feature_specs),
            psutil.cpu_count(logical=True),
        ),
        cache=DiskCache(
            feature_cache_dir=project_info.project_path / "feature_cache",
        ),
        drop_pred_times_with_insufficient_look_distance=drop_pred_times_with_insufficient_look_distance,
        predictor_col_name_prefix=project_info.prefix.predictor,
        outcome_col_name_prefix=project_info.prefix.outcome,
        timestamp_col_name=project_info.col_names.timestamp,
        entity_id_col_name=project_info.col_names.id,
    )

    flattened_dataset.add_age(
        date_of_birth_df=birthdays(),
        date_of_birth_col_name="date_of_birth",
    )

    flattened_dataset.add_spec(spec=feature_specs)

    return flattened_dataset.get_df()



"""Loaders for patient IDs."""

from typing import Optional

import pandas as pd

from psycop_feature_generation.loaders.raw.sql_load import sql_load


def load_ids(split: str, n_rows: Optional[int] = None) -> pd.DataFrame:
    """Loads ids for a given split.

    Args:
        split (str): Which split to load IDs from. Takes either "train", "test" or "val". # noqa: DAR102
        n_rows: Number of rows to return. Defaults to None.

    Returns:
        pd.DataFrame: Only dw_ek_borger column with ids
    """
    view = f"[psycop_{split}_ids]"

    sql = f"SELECT * FROM [fct].{view}"

    df = sql_load(sql, database="USR_PS_FORSK", chunksize=None, n_rows=n_rows)

    return df.reset_index(drop=True)



"""Setup for the project."""
import logging
import sys
import tempfile
import time
from pathlib import Path
from typing import Literal

import wandb
from timeseriesflattener.feature_spec_objects import (  # pylint: disable=no-name-in-module
    BaseModel,
)

from psycop_feature_generation.utils import RELATIVE_PROJECT_ROOT, SHARED_RESOURCES_PATH

log = logging.getLogger(__name__)


class Prefixes(BaseModel):
    """Prefixes for feature specs."""

    predictor: str = "pred"
    outcome: str = "outc"
    eval: str = "eval"


class ColNames(BaseModel):
    """Column names for feature specs."""

    timestamp = "timestamp"
    id = "dw_ek_borger"


class ProjectInfo(BaseModel):
    """Collection of project info."""

    project_name: str
    project_path: Path
    feature_set_path: Path
    feature_set_prefix: str
    dataset_format: Literal["parquet", "csv"] = "parquet"
    prefix: Prefixes = Prefixes()
    col_names: ColNames = ColNames()

    def __init__(self, **data):
        super().__init__(**data)

        # Iterate over each attribute. If the attribute is a Path, create it if it does not exist.
        for attr in self.__dict__:
            if isinstance(attr, Path):
                attr.mkdir(exist_ok=True, parents=True)


def create_feature_set_path(
    proj_path: Path,
    feature_set_id: str,
) -> Path:
    """Create save directory.

    Args:
        proj_path (Path): Path to project.
        feature_set_id (str): Feature set id.

    Returns:
        Path: Path to sub directory.
    """

    # Split and save to disk
    # Create directory to store all files related to this run
    save_dir = proj_path / "feature_sets" / feature_set_id

    save_dir.mkdir(exist_ok=True, parents=True)

    return save_dir


def get_project_info(
    project_name: str,
) -> ProjectInfo:
    """Setup for main.

    Args:
        project_name (str): Name of project.
    Returns:
        tuple[Path, str]: Tuple of project path, and feature_set_id
    """
    log.info("Setting up project")
    proj_path = SHARED_RESOURCES_PATH / project_name

    current_user = Path().home().name
    feature_set_id = f"psycop_{project_name}_{current_user}_features_{time.strftime('%Y_%m_%d_%H_%M')}"

    feature_set_path = create_feature_set_path(
        feature_set_id=feature_set_id,
        proj_path=proj_path,
    )

    return ProjectInfo(
        project_path=proj_path,
        feature_set_path=feature_set_path,
        feature_set_prefix=feature_set_id,
        project_name=project_name,
    )


def init_wandb(
    project_info: ProjectInfo,
) -> None:
    """Initialise wandb logging. Allows to use wandb to track progress, send
    Slack notifications if failing, and track logs.

    Args:
        project_info (ProjectInfo): Project info.
    """

    feature_settings = {
        "feature_set_path": project_info.feature_set_path,
    }

    # on Overtaci, the wandb tmp directory is not automatically created,
    # so we create it here.
    # create debug-cli.one folders in /tmp and project dir
    if sys.platform == "win32":
        (Path(tempfile.gettempdir()) / "debug-cli.onerm").mkdir(
            exist_ok=True,
            parents=True,
        )
        (RELATIVE_PROJECT_ROOT / "wandb" / "debug-cli.onerm").mkdir(
            exist_ok=True,
            parents=True,
        )

    wandb.init(
        project=f"{project_info.project_name}-feature-generation",
        config=feature_settings,
    )



"""Utilites for testing."""

from io import StringIO

import numpy as np
import pandas as pd
import pytest
from pandas import DataFrame

from psycop_feature_generation.loaders.synth.raw.load_synth_data import (
    load_synth_outcome,
    load_synth_prediction_times,
)
from psycop_feature_generation.utils import data_loaders


def convert_cols_with_matching_colnames_to_datetime(
    df: DataFrame,
    colname_substr: str,
) -> DataFrame:
    """Convert columns that contain colname_substr in their name to datetimes
    Args:
        df (DataFrame): The df to convert. # noqa: DAR101
        colname_substr (str): Substring to match on. # noqa: DAR101

    Returns:
        DataFrame: The converted df
    """
    df.loc[:, df.columns.str.contains(colname_substr)] = df.loc[
        :,
        df.columns.str.contains(colname_substr),
    ].apply(pd.to_datetime)

    return df


def str_to_df(
    string: str,
    convert_timestamp_to_datetime: bool = True,
    convert_np_nan_to_nan: bool = True,
    convert_str_to_float: bool = False,
    add_pred_time_uuid: bool = False,
    entity_id_colname: str = "entity_id",
    timestamp_col_name: str = "timestamp",
) -> DataFrame:
    """Convert a string representation of a dataframe to a dataframe.

    Args:
        string (str): A string representation of a dataframe.
        convert_timestamp_to_datetime (bool): Whether to convert the timestamp column to datetime. Defaults to True.
        convert_np_nan_to_nan (bool): Whether to convert np.nan to np.nan. Defaults to True.
        convert_str_to_float (bool): Whether to convert strings to floats. Defaults to False.
        add_pred_time_uuid (bool): Whether to infer a pred_time_uuid column from entity_id and timestamp columns. Defaults to False.
        entity_id_colname (str): The name of the entity_id column. Defaults to "entity_id".
        timestamp_col_name (str): The name of the timestamp column. Defaults to "timestamp".

    Returns:
        DataFrame: A dataframe.
    """
    # Drop comments for each line if any exist inside the str
    string = string[: string.rfind("#")]

    df = pd.read_table(StringIO(string), sep=",", index_col=False)

    if convert_timestamp_to_datetime:
        df = convert_cols_with_matching_colnames_to_datetime(df, "timestamp")

    if convert_np_nan_to_nan:
        # Convert "np.nan" str to the actual np.nan
        df = df.replace("np.nan", np.nan)

    if convert_str_to_float:
        # Convert all str to float
        df = df.apply(pd.to_numeric, axis=0, errors="coerce")

    if add_pred_time_uuid:
        df["pred_time_uuid"] = (
            df[entity_id_colname].astype(str) + "_" + df[timestamp_col_name].astype(str)
        )

    # Drop "Unnamed" cols
    return df.loc[:, ~df.columns.str.contains("^Unnamed")]


@data_loaders.register("load_event_times")
def load_event_times():
    """Load event times."""
    event_times_str = """dw_ek_borger,timestamp,value,
                    1,2021-12-30 00:00:01, 1
                    1,2021-12-29 00:00:02, 2
                    """

    return str_to_df(event_times_str)


def check_any_item_in_list_has_str(list_of_str: list, str_: str):
    """Check if any item in a list contains a string.

    Args:
        list_of_str (list): A list of strings.
        str_ (str): A string.

    Returns:
        bool: True if any item in the list contains the string.
    """
    return any(str_ in item for item in list_of_str)


@pytest.fixture(scope="function")
def synth_prediction_times():
    """Load the prediction times."""
    return load_synth_prediction_times()


@pytest.fixture(scope="function")
def synth_outcome():
    """Load the synth outcome times."""
    return load_synth_outcome()



"""Loaders for medications."""
# pylint: disable=missing-function-docstring
import logging
from typing import Optional, Union

import pandas as pd

from psycop_feature_generation.loaders.raw.utils import load_from_codes
from psycop_feature_generation.utils import data_loaders

log = logging.getLogger(__name__)


def load(
    atc_code: Union[str, list[str]],
    output_col_name: Optional[str] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    wildcard_code: Optional[bool] = True,
    n_rows: Optional[int] = None,
    exclude_atc_codes: Optional[list[str]] = None,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    """Load medications. Aggregates prescribed/administered if both true. If
    wildcard_atc_code, match from atc_code*. Aggregates all that match. Beware
    that data is incomplete prior to sep. 2016 for prescribed medications.

    Args:
        atc_code (str): ATC-code prefix to load. Matches atc_code_prefix*.
            Aggregates all.
        output_col_name (str, optional): Name of output_col_name. Contains 1 if
            atc_code matches atc_code_prefix, 0 if not.Defaults to
            {atc_code_prefix}_value.
        load_prescribed (bool, optional): Whether to load prescriptions. Defaults to
            False. Beware incomplete until sep 2016.
        load_administered (bool, optional): Whether to load administrations.
            Defaults to True.
        wildcard_code (bool, optional): Whether to match on atc_code* or
            atc_code.
        n_rows (int, optional): Number of rows to return. Defaults to None, in which case all rows are returned.
        exclude_atc_codes (list[str], optional): Drop rows if atc_code is a direct match to any of these. Defaults to None.
        administration_route (str, optional): Whether to subset by a specific administration route, e.g. 'OR', 'IM' or 'IV'. Only applicable for administered medication, not prescribed. Defaults to None.
        administration_method (str, optional): Whether to subset by method of administration, e.g. 'PN' or 'Fast'. Only applicable for administered medication, not prescribed. Defaults to None.

    Returns:
        pd.DataFrame: Cols: dw_ek_borger, timestamp, {atc_code_prefix}_value = 1
    """

    if load_prescribed and any([administration_method, administration_route]):
        raise TypeError(
            "load() got an unexpected combination of arguments. When load_prescribed=True, administration_method and administration_route must be NoneType objects.",
        )

    if load_prescribed:
        log.warning(
            "Beware, there are missing prescriptions until september 2016. "
            "Hereafter, data is complete. See the wiki (OBS: Medication) for more details.",
        )

    df = pd.DataFrame()

    if load_prescribed and load_administered:
        n_rows = int(n_rows / 2) if n_rows else None

    if load_prescribed:
        df_medication_prescribed = load_from_codes(
            codes_to_match=atc_code,
            code_col_name="atc",
            source_timestamp_col_name="datotid_ordinationstart",
            view="FOR_Medicin_ordineret_inkl_2021_feb2022",
            output_col_name=output_col_name,
            match_with_wildcard=wildcard_code,
            n_rows=n_rows,
            exclude_codes=exclude_atc_codes,
            load_diagnoses=False,
            administration_route=administration_route,
            administration_method=administration_method,
        )

        df = pd.concat([df, df_medication_prescribed])

    if load_administered:
        df_medication_administered = load_from_codes(
            codes_to_match=atc_code,
            code_col_name="atc",
            source_timestamp_col_name="datotid_administration_start",
            view="FOR_Medicin_administreret_inkl_2021_feb2022",
            output_col_name=output_col_name,
            match_with_wildcard=wildcard_code,
            n_rows=n_rows,
            exclude_codes=exclude_atc_codes,
            load_diagnoses=False,
            administration_route=administration_route,
            administration_method=administration_method,
        )
        df = pd.concat([df, df_medication_administered])

    if output_col_name is None:
        if isinstance(atc_code, list):
            # Joint list of atc_codes
            output_col_name = "_".join(atc_code)
        else:
            output_col_name = atc_code

    df.rename(
        columns={
            output_col_name: "value",
        },
        inplace=True,
    )

    return df.reset_index(drop=True).drop_duplicates(
        subset=["dw_ek_borger", "timestamp", "value"],
        keep="first",
    )


def concat_medications(
    output_col_name: str,
    atc_code_prefixes: list[str],
    n_rows: Optional[int] = None,
) -> pd.DataFrame:
    """Aggregate multiple blood_sample_ids (typically NPU-codes) into one
    column.

    Args:
        output_col_name (str): Name for new column.  # noqa: DAR102
        atc_code_prefixes (list[str]): list of atc_codes.
        n_rows (int, optional): Number of atc_codes to aggregate. Defaults to None.

    Returns:
        pd.DataFrame
    """
    dfs = [
        load(
            atc_code=f"{id}",
            output_col_name=output_col_name,
            n_rows=n_rows,
        )
        for id in atc_code_prefixes
    ]

    return (
        pd.concat(dfs, axis=0)
        .drop_duplicates(
            subset=["dw_ek_borger", "timestamp", "value"],
            keep="first",
        )
        .reset_index(drop=True)
    )


# data_loaders primarly used in psychiatry
@data_loaders.register("antipsychotics")
def antipsychotics(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    """All antipsyhotics, except Lithium.

    Lithium is typically considered a mood stabilizer, not an
    antipsychotic.
    """
    return load(
        atc_code="N05A",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
        exclude_atc_codes=["N05AN01"],
    )


# 1. generation antipsychotics [flupentixol, pimozid, haloperidol, zuclopenthixol, melperon,pipamperon, chlorprotixen]
@data_loaders.register("first_gen_antipsychotics")
def first_gen_antipsychotics(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code=[
            "N05AF01",
            "N05AG02",
            "N05AD01",
            "N05AF05",
            "N05AD03",
            "N05AD05",
            "N05AF03",
        ],
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=False,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


# 2. generation antipsychotics [amisulpride, aripiprazole,asenapine, brexpiprazole, cariprazine, lurasidone, olanzapine, paliperidone, Quetiapine, risperidone, sertindol]
@data_loaders.register("second_gen_antipsychotics")
def second_gen_antipsychotics(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code=[
            "N05AL05",
            "N05AX12",
            "N05AH05",
            "N05AX16",
            "N05AX15",
            "N05AE02",
            "N05AE05",
            "N05AH03",
            "N05AX13",
            "N05AH04",
            "N05AX08",
            "N05AE04",
            "N05AE03",
        ],
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=False,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("top_10_weight_gaining_antipsychotics")
def top_10_weight_gaining_antipsychotics(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    """Top 10 weight gaining antipsychotics based on Huhn et al.

    2019. Only 5 of them are marketed in Denmark.
    """
    return load(
        atc_code=[
            "N05AH03",
            "N05AE03",
            "N05AH04",
            "N05AX13",
            "N05AX08",
        ],
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=False,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("olanzapine")
def olanzapine(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="N05AH03",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=False,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("clozapine")
def clozapine(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="N05AH02",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=False,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("anxiolytics")
def anxiolytics(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="N05B",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("benzodiazepines")
def benzodiazepines(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="N05BA",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("benzodiazepine_related_sleeping_agents")
def benzodiazepine_related_sleeping_agents(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code=["N05CF01", "N05CF02"],
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("pregabaline")
def pregabaline(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="N03AX16",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("opioid_dependence")
def opioid_dependence(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    """All opioid dependence medications."""
    return load(
        atc_code="N07BC",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("buprenorphine")
def buprenorphine(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    """Opioid dependence medications with the active ingredient
    buprenorphine."""
    return load(
        atc_code="N07BC01",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=False,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("methadone")
def methadone(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    """Opioid dependence medications with the active ingredient methadone."""
    return load(
        atc_code="N07BC02",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=False,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("naxolone")
def naxolone(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    """Opioid dependence medications with the active ingredients naxolone and
    buprenorphine."""
    return load(
        atc_code="N07BC51",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=False,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("hypnotics and sedatives")
def hypnotics(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="N05C",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("antidepressives")
def antidepressives(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="N06A",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


# SSRIs [escitalopram, citalopram, fluvoxamin, fluoxetin, paroxetin]
@data_loaders.register("ssri")
def ssri(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="N06AB",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


# SNRIs [duloxetin, venlafaxin]
@data_loaders.register("snri")
def snri(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code=["N06AX21", "N06AX16"],
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=False,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


# TCAs
@data_loaders.register("tca")
def tca(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="N06AA",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("selected_nassa")
def selected_nassa(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code=["N06AX11", "N06AX03"],
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("lithium")
def lithium(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="N05AN01",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=False,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("valproate")
def valproate(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="N03AG01",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=False,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("lamotrigine")
def lamotrigine(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="N03AX09",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=False,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("hyperactive disorders medications")
def hyperactive_disorders_medications(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="N06B",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("dementia medications")
def dementia_medications(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="N06D",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("anti-epileptics")
def anti_epileptics(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="N03",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


# medications used in alcohol abstinence treatment [thiamin, b-combin, klopoxid, fenemal]
@data_loaders.register("alcohol_abstinence")
def alcohol_abstinence(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code=["A11DA01", "A11EA", "N05BA02", "N03AA02"],
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=False,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


# data loaders for medications primarily used outside psychiatry
@data_loaders.register("alimentary_tract_and_metabolism_medications")
def alimentary_medications(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="A",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("blood_and_blood_forming_organs_medications")
def blood_medications(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="B",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("cardiovascular_medications")
def cardiovascular_medications(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="C",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("dermatologicals")
def dermatological_medications(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="D",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("genito_urinary_system_and_sex_hormones_medications")
def genito_sex_medications(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="G",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("systemic_hormonal_preparations")
def hormonal_medications(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="H",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("antiinfectives")
def antiinfectives(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="J",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("antineoplastic")
def antineoplastic(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="L",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("musculoskeletal_medications")
def musculoskeletal_medications(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="M",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("nervous_system_medications")
def nervous_system_medications(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="N",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("analgesics")
def analgesic(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="N02",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("antiparasitic")
def antiparasitic(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="P",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("respiratory_medications")
def respiratory_medications(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="R",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("sensory_organs_medications")
def sensory_medications(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="S",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("various_medications")
def various_medications(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="V",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("statins")
def statins(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="C10AA",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("antihypertensives")
def antihypertensives(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="C02",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("diuretics")
def diuretics(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    return load(
        atc_code="C07",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )


@data_loaders.register("gerd_drugs")
def gerd_drugs(
    n_rows: Optional[int] = None,
    load_prescribed: Optional[bool] = False,
    load_administered: Optional[bool] = True,
    administration_route: Optional[str] = None,
    administration_method: Optional[str] = None,
) -> pd.DataFrame:
    """Gastroesophageal reflux disease (GERD) drugs."""
    return load(
        atc_code="A02",
        load_prescribed=load_prescribed,
        load_administered=load_administered,
        wildcard_code=True,
        n_rows=n_rows,
        administration_route=administration_route,
        administration_method=administration_method,
    )



"""Loaders for synth data."""

from typing import Optional

import pandas as pd

from psycop_feature_generation.utils import RELATIVE_PROJECT_ROOT, data_loaders


def load_raw_test_csv(filename: str, n_rows: Optional[int] = None) -> pd.DataFrame:
    """Load raw csv.

    Args:
        filename (str): Name of the file to load.
        n_rows (int, optional): Number of rows to load. Defaults to None.
    """
    df = pd.read_csv(
        RELATIVE_PROJECT_ROOT / "tests" / "test_data" / "raw" / filename,
        nrows=n_rows,
    )

    # Convert timestamp col to datetime
    if "timestamp" in df.columns:
        df["timestamp"] = pd.to_datetime(df["timestamp"])

    return df


@data_loaders.register("synth_predictor_float")
def synth_predictor_float(
    n_rows: Optional[int] = None,
) -> pd.DataFrame:
    """Load synth predictor data.".

    Args:
        n_rows: Number of rows to return. Defaults to None which returns entire coercion data view.

    Returns:
        pd.DataFrame
    """
    return load_raw_test_csv("synth_raw_float_1.csv", n_rows=n_rows)


@data_loaders.register("synth_predictor_binary")
def synth_predictor_binary(
    n_rows: Optional[int] = None,
) -> pd.DataFrame:
    """Load synth predictor data.".

    Args:
        n_rows: Number of rows to return. Defaults to None which returns entire coercion data view.

    Returns:
        pd.DataFrame
    """
    return load_raw_test_csv("synth_raw_binary_1.csv", n_rows=n_rows)


@data_loaders.register("synth_outcome")
def load_synth_outcome(
    n_rows: Optional[int] = None,
) -> pd.DataFrame:
    """Load synth predictor data.".

    Args:
        n_rows: Number of rows to return. Defaults to None which returns entire coercion data view.

    Returns:
        pd.DataFrame
    """
    # Get first row for each id
    df = load_raw_test_csv("synth_raw_binary_2.csv", n_rows=n_rows)
    df = df.groupby("dw_ek_borger").last().reset_index()
    return df


@data_loaders.register("synth_prediction_times")
def load_synth_prediction_times(
    n_rows: Optional[int] = None,
) -> pd.DataFrame:
    """Load synth predictor data.".

    Args:
        n_rows: Number of rows to return. Defaults to None which returns entire coercion data view.

    Returns:
        pd.DataFrame
    """
    return load_raw_test_csv("synth_prediction_times.csv", n_rows=n_rows)



"""Exemplifies how to debug an SQL loader function.

Two primary purposes:
1. Check the possible values the dataframe can contain and
2. Check that the dataframe conforms to the required format.
"""

from typing import Any

import pandas as pd

import psycop_feature_generation.loaders.raw as raw_loaders
from psycop_feature_generation.data_checks.raw.check_predictor_lists import check_raw_df


def will_it_float(value: Any) -> bool:
    """Check if a value can be converted to a float.

    Args:
        value (Any): A value.

    Returns:
        bool: True if the value can be converted to a float, False otherwise.
    """
    try:
        float(value)
        return True
    except ValueError:
        return False


def get_prop_of_each_unique_value_for_non_floats(series: pd.Series) -> pd.Series:
    """Get the proportion of each unique value in a series, but only for value
    which cannot be converted to floats.

    Args:
        series (pd.Series): A pandas series.

    Returns:
        pd.Series: A series with the proportion of each unique value in the
        original series.
    """
    if series.dtype in ["float64", "int64"]:
        return "All values in series can be converted to floats."

    # Find all strings that start with a number
    starts_with_number_idx = series.str.match(r"^\d+").fillna(False)

    # Convert all strings that start with a number to floats
    # Replace all "," with "."
    series[starts_with_number_idx] = series[starts_with_number_idx].str.replace(
        ",",
        ".",
    )

    # Convert all str in series to float
    series[starts_with_number_idx] = series[starts_with_number_idx].astype(float)

    # Get the unique values
    unique_values = series.unique()

    # Get the proportion of each unique value
    prop_of_each_unique_value = series.value_counts(
        normalize=True,
    )

    # Get the unique values which cannot be converted to floats
    non_float_unique_values = [
        value
        for value in unique_values
        if (not will_it_float(value) and value is not None)
    ]

    # Get the proportion of each unique value which cannot be converted to floats
    prop_of_each_non_float_unique_value = prop_of_each_unique_value[
        non_float_unique_values
    ]

    return prop_of_each_non_float_unique_value


if __name__ == "__main__":
    df = raw_loaders.load_lab_results.ldl(
        n=1_000,
        values_to_load="numerical_and_coerce",
    )

    value_props = get_prop_of_each_unique_value_for_non_floats(df["value"])
    print(value_props)

    errors, duplicates = check_raw_df(
        df=df,
        required_columns=["dw_ek_borger", "timestamp", "value"],
        subset_duplicates_columns=["dw_ek_borger", "timestamp", "value"],
    )
    print(errors)
